"""
    Elementary IR relations.
"""
import sys

from . import ir, factory as f
from . import types

from typing import Optional

#
# Relations
#

# Comparators
def _comparator(name: str, input=True):
    overloads = [
        f.relation(name, [f.field("a", type, input), f.field("b", type, input)])
        for type in [types.Int, types.Float, types.Decimal64, types.Decimal128, types.String, types.Date, types.DateTime, types.EntityTypeVar]
    ]
    return f.relation(name, [f.field("a", types.Any, input), f.field("b", types.Any, input)], overloads=overloads)

gt = _comparator(">")
gte = _comparator(">=")
lt = _comparator("<")
lte = _comparator("<=")
neq = _comparator("!=")
eq = _comparator("=", False)

# Arithmetic operators
def _binary_op(name: str, with_string=False, result_type: Optional[ir.Type]=None):
    overload_types = [types.Int, types.Float, types.Decimal64, types.Decimal128]
    if with_string:
        overload_types.append(types.String)
    overloads = [
        f.relation(name, [
            f.input_field("a", type),
            f.input_field("b", type),
            f.field("c", result_type if result_type is not None else type)])
        for type in overload_types
    ]

    if with_string:
        return f.relation(name, [f.input_field("a", types.Any), f.input_field("b", types.Any), f.field("c", types.Any)], overloads=overloads)
    else:
        # If strings isn't added, then we're guaranteed to only have number overloads
        result_type = result_type if result_type is not None else types.Number
        return f.relation(name, [f.input_field("a", types.Number), f.input_field("b", types.Number), f.field("c", result_type)], overloads=overloads)

plus = _binary_op("+", with_string=True)
minus = _binary_op("-")
mul = _binary_op("*")
div = f.relation(
    "/",
    [f.input_field("a", types.Number), f.input_field("b", types.Number), f.field("c", types.Number)],
    overloads=[
        f.relation("/", [f.input_field("a", types.Int), f.input_field("b", types.Int), f.field("c", types.Float)]),
        f.relation("/", [f.input_field("a", types.Float), f.input_field("b", types.Float), f.field("c", types.Float)]),
        f.relation("/", [f.input_field("a", types.Decimal64), f.input_field("b", types.Decimal64), f.field("c", types.Decimal64)]),
        f.relation("/", [f.input_field("a", types.Decimal128), f.input_field("b", types.Decimal128), f.field("c", types.Decimal128)]),
    ],
)
mod = _binary_op("%")

trunc_div = f.relation("//", [f.input_field("a", types.Int), f.input_field("b", types.Int), f.field("c", types.Int)])

abs = f.relation(
    "abs",
    [f.input_field("a", types.Number), f.field("b", types.Number)],
    overloads=[
        f.relation("abs", [f.input_field("a", types.Int), f.field("b", types.Int)]),
        f.relation("abs", [f.input_field("a", types.Float), f.field("b", types.Float)]),
        f.relation("abs", [f.input_field("a", types.Decimal64), f.field("b", types.Decimal64)]),
        f.relation("abs", [f.input_field("a", types.Decimal128), f.field("b", types.Decimal128)]),
    ],
)

natural_log = f.relation(
    "natural_log",
    [f.input_field("a", types.Number), f.field("b", types.Number)],
    overloads=[
        f.relation("natural_log", [f.input_field("a", types.Int), f.field("b", types.Float)]),
        f.relation("natural_log", [f.input_field("a", types.Float), f.field("b", types.Float)]),
        f.relation("natural_log", [f.input_field("a", types.Decimal64), f.field("b", types.Decimal64)]),
        f.relation("natural_log", [f.input_field("a", types.Decimal128), f.field("b", types.Decimal128)]),
    ],
)

# Strings
concat = f.relation("concat", [f.input_field("a", types.String), f.input_field("b", types.String), f.field("c", types.String)])
num_chars = f.relation("num_chars", [f.input_field("a", types.String), f.field("b", types.Int)])
starts_with = f.relation("starts_with", [f.input_field("a", types.String), f.input_field("b", types.String)])
ends_with = f.relation("ends_with", [f.input_field("a", types.String), f.input_field("b", types.String)])
contains = f.relation("contains", [f.input_field("a", types.String), f.input_field("b", types.String)])
substring = f.relation("substring", [f.input_field("a", types.String), f.input_field("b", types.Int), f.input_field("c", types.Int), f.field("d", types.String)])
like_match = f.relation("like_match", [f.input_field("a", types.String), f.field("b", types.String)])

# Dates
date_year = f.relation("date_year", [f.input_field("a", types.Date), f.field("b", types.Int)])
date_month = f.relation("date_month", [f.input_field("a", types.Date), f.field("b", types.Int)])
date_day = f.relation("date_day", [f.input_field("a", types.Date), f.field("b", types.Int)])

# Conversions
int_to_float = f.relation("int_to_float", [f.input_field("a", types.Int), f.field("b", types.Float)])
int_to_decimal64 = f.relation("int_to_decimal64", [f.input_field("a", types.Int), f.field("b", types.Decimal64)])
int_to_decimal128 = f.relation("int_to_decimal128", [f.input_field("a", types.Int), f.field("b", types.Decimal128)])
float_to_int = f.relation("float_to_int", [f.input_field("a", types.Float), f.field("b", types.Int)])
float_to_decimal64 = f.relation("float_to_decimal64", [f.input_field("a", types.Float), f.field("b", types.Decimal64)])
float_to_decimal128 = f.relation("float_to_decimal128", [f.input_field("a", types.Float), f.field("b", types.Decimal128)])
decimal64_to_int = f.relation("decimal64_to_int", [f.input_field("a", types.Decimal64), f.field("b", types.Int)])
decimal64_to_float = f.relation("decimal64_to_float", [f.input_field("a", types.Decimal64), f.field("b", types.Float)])
decimal64_to_decimal128 = f.relation("decimal64_to_decimal128", [f.input_field("a", types.Decimal64), f.field("b", types.Decimal128)])
decimal128_to_int = f.relation("decimal128_to_int", [f.input_field("a", types.Decimal128), f.field("b", types.Int)])
decimal128_to_float = f.relation("decimal128_to_float", [f.input_field("a", types.Decimal128), f.field("b", types.Float)])
decimal128_to_decimal64 = f.relation("decimal128_to_decimal64", [f.input_field("a", types.Decimal128), f.field("b", types.Decimal64)])

# Other
range = f.relation("range", [
    f.input_field("start", types.Int),
    f.input_field("stop", types.Int),
    f.input_field("step", types.Int),
    f.field("result", types.Int),
])

hash = f.relation("rel_primitive_hash_tuple", [f.input_field("args", types.AnyList), f.field("hash", types.Hash)])
hash_128 = f.relation("hash", [f.input_field("args", types.AnyList), f.field("hash", types.Int)])

uuid_to_string = f.relation("uuid_to_string", [f.input_field("a", types.Int), f.field("b", types.String)])

rel_primitive_solverlib_fo_appl = f.relation("rel_primitive_solverlib_fo_appl", [
    f.input_field("op", types.Int),
    f.input_field("args", types.AnyList),
    f.field("result", types.String),
])

# Raw source code to be attached to the transaction, when the backend understands this language
raw_source = f.relation("raw_source", [f.input_field("lang", types.String), f.input_field("source", types.String)])

#
# Annotations
#

# indicates a relation is external to the system and, thus, backends should not rename or
# otherwise modify it
external = f.relation("external", [])
external_annotation = f.annotation(external, [])

# indicates an output is meant to be exported
export = f.relation("export", [])
export_annotation = f.annotation(export, [])

# indicates this relation is a concept population
concept_population = f.relation("concept_population", [])
concept_relation_annotation = f.annotation(concept_population, [])

# indicates this relation came in from CDC and will need to be shredded in Rel
from_cdc = f.relation("from_cdc", [])
from_cdc_annotation = f.annotation(from_cdc, [])

# indicates this relation is a type_box operation for value types
type_box = f.relation("type_box", [])
type_box_annotation = f.annotation(type_box, [])

#
# Aggregations
#
def aggregation(name: str, params: list[ir.Field], overload_types: Optional[list[tuple[ir.Type, ...]]] = None):
    """Defines an aggregation, which is a Relation whose first 2 fields are a projection
    and a group, followed by the params."""
    fields = params
    overloads = []
    if overload_types:
        param_sets = []
        for ts in overload_types:
            param_sets.append([ir.Field(param.name, t, param.input) for param, t in zip(params, ts)])
        overloads = [
            aggregation(name, typed_params, overload_types=None)
            for typed_params in param_sets
        ]
    return f.relation(name, fields, overloads=overloads)

# concat = aggregation("concat", [
#     f.input_field("sep", types.String),
#     f.input_field("over", types.StringSet),
#     f.field("result", types.String)
# ])
# note that count does not need "over" because it counts the projection
count = aggregation("count", [
    f.field("result", types.Int)
])
stats = aggregation("stats", [
    f.input_field("over", types.Number),
    f.field("std_dev", types.Number),
    f.field("mean", types.Number),
    f.field("median", types.Number),
])
sum = aggregation("sum", [
    f.input_field("over", types.Number),
    f.field("result", types.Number)
], overload_types=[
    (types.Int, types.Int),
    (types.Float, types.Float),
    (types.Decimal64, types.Decimal64),
    (types.Decimal128, types.Decimal128),
])
avg = aggregation("avg", [
    f.input_field("over", types.Number),
    f.field("result", types.Number)
], overload_types=[
    (types.Int, types.Float), # nb. Float because Int / Int is Float
    (types.Float, types.Float),
    (types.Decimal64, types.Decimal64),
    (types.Decimal128, types.Decimal128),
])
max = aggregation("max", [
    f.input_field("over", types.Any),
    f.field("result", types.Any)
], overload_types=[
    (types.Int, types.Int),
    (types.Float, types.Float),
    (types.Decimal64, types.Decimal64),
    (types.Decimal128, types.Decimal128),
    (types.String, types.String),
    (types.Date, types.Date),
    (types.DateTime, types.DateTime),
    (types.EntityTypeVar, types.EntityTypeVar),
])
min = aggregation("min", [
    f.input_field("over", types.Any),
    f.field("result", types.Any)
], overload_types=[
    (types.Int, types.Int),
    (types.Float, types.Float),
    (types.Decimal64, types.Decimal64),
    (types.Decimal128, types.Decimal128),
    (types.String, types.String),
    (types.Date, types.Date),
    (types.DateTime, types.DateTime),
    (types.EntityTypeVar, types.EntityTypeVar),
])

rel_primitive_solverlib_ho_appl = aggregation("rel_primitive_solverlib_ho_appl", [
    f.input_field("over", types.Any),
    f.field("op", types.Int),
    f.field("result", types.String),
])


# TODO: these are Rel specific, should be moved from here
# Conversions
string = f.relation("string", [f.input_field("a", types.Any), f.field("b", types.String)])
parse_date = f.relation("parse_date", [f.input_field("a", types.String), f.input_field("b", types.String), f.field("c", types.Date)])
parse_datetime = f.relation("parse_datetime", [f.input_field("a", types.String), f.input_field("b", types.String), f.field("c", types.DateTime)])
parse_decimal64 = f.relation("parse_decimal64", [f.input_field("a", types.String), f.field("b", types.Decimal64)])
parse_decimal128 = f.relation("parse_decimal128", [f.input_field("a", types.String), f.field("b", types.Decimal128)])
# Alias parse_decimal128 to parse_decimal
parse_decimal = parse_decimal128

# Implicit decimal conversions
# These allow writing decimal128(1) in QB and letting the typer insert the conversion from any number type
decimal64 = f.relation("decimal64", [f.input_field("input", types.Decimal64), f.field("result", types.Decimal64)])
decimal128 = f.relation("decimal128", [f.input_field("input", types.Decimal128), f.field("result", types.Decimal128)])


# Date construction with less overhead
construct_date = f.relation("construct_date", [f.input_field("year", types.Int), f.input_field("month", types.Int), f.input_field("day", types.Int), f.field("date", types.Date)])
construct_datetime = f.relation("construct_datetime", [f.input_field("year", types.Int), f.input_field("month", types.Int), f.input_field("day", types.Int), f.input_field("hour", types.Int), f.input_field("minute", types.Int), f.input_field("second", types.Int), f.field("datetime", types.DateTime)])
construct_datetime_ms_tz = f.relation("construct_datetime_ms_tz", [f.input_field("year", types.Int), f.input_field("month", types.Int), f.input_field("day", types.Int), f.input_field("hour", types.Int), f.input_field("minute", types.Int), f.input_field("second", types.Int), f.input_field("milliseconds", types.Int), f.input_field("timezone", types.String), f.field("datetime", types.DateTime)])

#
# Public access to built-in relations
#

def is_builtin(r: ir.Relation):
    return r in builtin_relations or r in builtin_overloads

def is_annotation(r: ir.Relation):
    return r in builtin_annotations

def _compute_builtin_relations() -> list[ir.Relation]:
    module = sys.modules[__name__]
    relations = []
    for name in dir(module):
        builtin = getattr(module, name)
        if isinstance(builtin, ir.Relation) and builtin not in builtin_annotations:
            relations.append(builtin)
    return relations

def _compute_builtin_overloads() -> list[ir.Relation]:
    module = sys.modules[__name__]
    overloads = []
    for name in dir(module):
        builtin = getattr(module, name)
        if isinstance(builtin, ir.Relation) and builtin not in builtin_annotations:
            if builtin.overloads:
                for overload in builtin.overloads:
                    if overload not in builtin_annotations:
                        overloads.append(overload)
    return overloads

# manually maintain the list of relations that are actually annotations
builtin_annotations = [external, export, concept_population, from_cdc, type_box]
builtin_annotations_by_name = dict((r.name, r) for r in builtin_annotations)

builtin_relations = _compute_builtin_relations()
builtin_overloads = _compute_builtin_overloads()
builtin_relations_by_name = dict((r.name, r) for r in builtin_relations)

conversion_builtins = [string, decimal64, decimal128, int_to_float, int_to_decimal64, int_to_decimal128, float_to_int,
                              float_to_decimal64, float_to_decimal128, decimal64_to_int, decimal64_to_float,
                              decimal64_to_decimal128, decimal128_to_int, decimal128_to_float, decimal128_to_decimal64]

string_binary_builtins = [num_chars, starts_with, ends_with, contains, like_match]



from abc import ABC, abstractmethod
from enum import Enum
from typing import Sequence, Optional, TypeVar, Generic, Union

from relationalai.early_access.dsl.core.utils import generate_stable_uuid
from relationalai.early_access.dsl.orm.relationships import Role
from relationalai.early_access.dsl.orm.types import Concept
from relationalai.early_access.metamodel.util import OrderedSet, ordered_set


class Constraint(ABC):
    @abstractmethod
    def _validate(self, ontology) -> bool:
        """Validate the constraint against the ontology."""

    @abstractmethod
    def _unique_name(self) -> str:
        """Validate the constraint against the ontology."""

    @abstractmethod
    def _desugar(self):
        """Declare QB constraints"""

    def _guid(self):
        return generate_stable_uuid(self._unique_name())

    def __eq__(self, other):
        if isinstance(other, Constraint):
            return self._guid() == other._guid()
        return False

    def __hash__(self):
        return hash(self._guid())


class RoleConstraint(Constraint):

    def roles(self) -> list[Role]:
        return self._roles

    def __init__(self, roles: list[Role]):
        self._roles = roles
        self._name = f'{type(self).__name__}{"".join(role._guid() for role in self._roles)}'

    def _unique_name(self) -> str:
        return self._name


class Unique(RoleConstraint):
    def __init__(self, *roles: Role, is_preferred_identifier=False):
        super().__init__(list(roles))
        self.is_preferred_identifier = is_preferred_identifier

    def _validate(self, ontology) -> bool:
        # Implement uniqueness validation logic here
        return True

    def _desugar(self):
        # todo: Implement QB representation
        pass

    def _is_internal(self) -> bool:
        # only when there is a single role
        return len(self._roles) == 1

    def __repr__(self):
        return f'Unique(({", ".join(role._guid() for role in self._roles)}), preferred_identifier={self.is_preferred_identifier})'


class Mandatory(RoleConstraint):
    def __init__(self, role: Role):
        super().__init__([role])

    def _validate(self, ontology) -> bool:
        # Implement mandatory validation logic here
        return True

    def _desugar(self):
        # todo: Implement QB representation
        pass

    def __repr__(self):
        return f'Mandatory({self._roles[0]._guid()})'


class InclusiveRoleConstraint(RoleConstraint):
    def __init__(self, *roles: Role):
        super().__init__(list(roles))

    def _validate(self, ontology) -> bool:
        # Implement inclusive validation logic here
        return True

    def _desugar(self):
        # todo: Implement QB representation
        pass

    def __repr__(self):
        return f'Inclusive(({", ".join(role._guid() for role in self._roles)})'


class ExclusiveRoleConstraint(RoleConstraint):
    def __init__(self, *roles: Role):
        super().__init__(list(roles))

    def _validate(self, ontology) -> bool:
        # Implement exclusive validation logic here
        return True

    def _desugar(self):
        # todo: Implement QB representation
        pass

    def __repr__(self):
        return f'Exclusive(({", ".join(role._guid() for role in self._roles)})'


class RingType(Enum):
    IRREFLEXIVE = 1
    ANTISYMMETRIC = 2
    ASYMMETRIC = 3
    STRONGLY_INTRANSITIVE = 4
    INTRANSITIVE = 5
    ACYCLIC = 6
    PURELY_REFLEXIVE = 7
    REFLEXIVE = 8
    SYMMETRIC = 9
    TRANSITIVE = 10


class RingConstraint(RoleConstraint):
    def __init__(self, constraint_types: list[RingType], *roles: Role):
        super().__init__(list(roles))
        self.types = constraint_types

    def _validate(self, ontology) -> bool:
        # Implement exclusive ring constraint logic here
        return True

    def _desugar(self):
        # todo: Implement QB representation
        pass

    def __repr__(self):
        return f'Ring(({", ".join(str(type) for type in self.types)}), ({", ".join(role._guid() for role in self._roles)}))'


# Role value constraint
T = TypeVar('T', int, float, str)


class Range(Generic[T]):

    def __init__(self, start: Optional[T], end: Optional[T]):
        if start is None and end is None:
            raise ValueError("'start' and 'end' cannot be None")
        if start is not None and end is not None:
            if type(start) is not type(end):
                raise TypeError("'start' and 'end' must be of same type")
            if start > end:  # type: ignore[reportOperatorIssue]
                raise ValueError("'start' must be less than 'end'")
        self._start = start
        self._end = end

    def _matches(self, value: T) -> bool:
        if self._start is not None and value < self._start:
            return False
        if self._end is not None and value > self._end:
            return False
        return True

    def _type(self):
        if self._start is not None:
            return type(self._start)
        else:
            return type(self._end)

    def __repr__(self):
        return f"Range({self._start}, {self._end})"

    @staticmethod
    def between(start: T, end: T):
        return Range(start, end)

    @staticmethod
    def to_value(value: T):
        return Range(None, value)

    @staticmethod
    def from_value(value: T):
        return Range(value, None)


class RoleValueConstraint(RoleConstraint, Generic[T]):
    def __init__(self, role: Role, values: Sequence[Union[T, Range[T]]]):
        super().__init__([role])
        self._values = values

    def values(self):
        return self._values

    def _validate(self, ontology) -> bool:
        # Implement role value validation logic here
        return True

    def _desugar(self):
        # todo: Implement QB representation
        pass

    def __repr__(self):
        return f'RoleValueConstraint({self._roles[0]._guid()}, values={self._values})'


class SubtypeConstraint(Constraint):
    _concepts: OrderedSet[Concept]

    def __init__(self, *concepts: Concept):
        self._concepts = ordered_set(*concepts)
        if len(concepts) < 2:
            raise ValueError("Invalid subtype constraint. A constraint should hold at least 2 concepts")
        first = next(iter(self._concepts))
        for c in self._concepts:
            if len(c._extends) == 0:
                raise ValueError(f"Invalid subtype constraint. '{c}' is not a subtype")
            if first._extends != c._extends:
                raise ValueError(f"Invalid subtype constraint. '{first}' and '{c}' must have the same parents")
        self._name = f'{type(self).__name__}({", ".join(str(c) for c in sorted(self._concepts, key=lambda c: str(c)))})'

    def _unique_name(self) -> str:
        return self._name

    def concepts(self) -> OrderedSet[Concept]:
        return self._concepts

    def __repr__(self):
        return self._name


class ExclusiveSubtypeConstraint(SubtypeConstraint):

    def __init__(self, *concepts: Concept):
        super().__init__(*concepts)

    def _validate(self, ontology) -> bool:
        # Implement validation logic here
        return True

    def _desugar(self):
        # todo: Implement QB representation
        pass


class InclusiveSubtypeConstraint(SubtypeConstraint):

    def __init__(self, *concepts: Concept):
        super().__init__(*concepts)

    def _validate(self, ontology) -> bool:
        # Implement validation logic here
        return True

    def _desugar(self):
        # todo: Implement QB representation
        pass



from collections import defaultdict

from relationalai.early_access.dsl.adapters.orm.model import ExclusiveInclusiveSubtypeFact, ExclusiveSubtypeFact
from relationalai.early_access.dsl.adapters.orm.parser import ORMParser
from relationalai.early_access.dsl.core.types.standard import Integer, UnsignedInteger, String, DateTime, Date, Decimal
from relationalai.early_access.dsl.ontologies.models import Model
from relationalai.early_access.dsl.types.entities import AbstractEntityType


class ORMAdapter:

    def __init__(self, orm_file_path: str):
        self._parser = ORMParser(orm_file_path)
        self._relationship_role_value_constraints = defaultdict()
        self.model = self.orm_to_model()

    def orm_to_model(self):
        model = Model(self._parser.model_name())

        self._add_value_types(model)
        self._add_entity_types(model)
        self._add_subtype_relationships(model)
        self._add_relationships(model)
        self._add_external_identifying_relationships(model)
        self._add_role_value_constraints(model)

        return model

    def _add_value_types(self, md):
        for vt in self._parser.value_types().values():
            md.value_type(vt.name, self._map_datatype(vt.data_type))

    def _add_entity_types(self, md):
        for et in self._parser.entity_types().values():
            md.entity_type(et.name, ref_mode=et.ref_mode)

    def _add_subtype_relationships(self, md):
        for parent, children in self._parser.subtype_facts().items():
            supertype = md.lookup_concept(parent)
            subtypes = []
            exclusives = []
            exclusives_and_inclusives = []
            for child in children:
                sub = md.lookup_concept(child.subtype_name)
                if isinstance(child, ExclusiveInclusiveSubtypeFact):
                    exclusives_and_inclusives.append(sub)
                elif isinstance(child, ExclusiveSubtypeFact):
                    exclusives.append(sub)
                else:
                    subtypes.append(sub)
            if len(exclusives_and_inclusives) > 0:
                md.subtype_arrow(supertype, exclusives_and_inclusives, exclusive=True, inclusive=True)
            if len(exclusives) > 0:
                md.subtype_arrow(supertype, exclusives, exclusive=True)
            if len(subtypes) > 0:
                md.subtype_arrow(supertype, subtypes)

    def _add_relationships(self, md):
        object_types = self._parser.object_types()
        unique_roles = self._parser.unique_roles()
        mandatory_roles = self._parser.mandatory_roles()
        role_value_constraints = self._parser.role_value_constraints()
        fact_type_to_roles = self._parser.fact_type_to_roles()
        fact_type_to_internal_ucs = self._parser.fact_type_to_internal_ucs()
        fact_type_to_complex_ucs = self._parser.fact_type_to_complex_ucs()
        identifier_fact_type_to_entity_type = self._parser.identifier_fact_type_to_entity_type()
        for fact_type, reading_orders in self._parser.fact_type_readings().items():
            with md.relationship() as rel:
                role_idx_to_player = list()
                # Use the first reading to add the players
                for role in fact_type_to_roles[fact_type]:
                    role_name = role.name or None
                    role_unique = role.id in unique_roles
                    role_mandatory = role.id in mandatory_roles
                    p = md.lookup_concept(object_types[role.player].name)
                    rel.role(p, name=role_name, unique=role_unique, mandatory=role_mandatory)
                    role_idx_to_player.append(p)
                    if role.id in role_value_constraints:
                        self._relationship_role_value_constraints[rel] = role_value_constraints[role.id]
                # Create the readings
                for rdo in reading_orders:
                    argz = []
                    for rdo_role in rdo.roles:
                        orm_role = rdo_role.role
                        player_id = orm_role.player
                        player_name = object_types[player_id].name
                        player = md.lookup_concept(player_name)
                        role_idx = role_idx_to_player.index(player)
                        role = rel.role_at(role_idx)
                        argz.append(role)
                        if rdo_role.text:
                            argz.append(rdo_role.text)
                        if rdo_role.prefix or rdo_role.postfix:
                            role.verbalization(prefix=rdo_role.prefix, postfix=rdo_role.postfix)
                    entity_type = identifier_fact_type_to_entity_type.get(fact_type)
                    if entity_type:
                        player = md.lookup_concept(entity_type.name)
                        if argz[0].player_type == player:
                            ref_mode = entity_type.ref_mode
                            rel.relation(*argz, name=ref_mode) if ref_mode else rel.relation(*argz)
                        else:
                            rel.relation(*argz)
                    else:
                        rel.relation(*argz)
                # Marking identifying relationships
                if fact_type in fact_type_to_internal_ucs:
                    for uc in fact_type_to_internal_ucs[fact_type]:
                        if uc.identifies:
                            et = md.lookup_concept(object_types[uc.identifies].name)
                            role_name = role_idx_to_player.index(et)
                            md.ref_scheme(rel.relations()[role_name])
                # Add constraint spanning over multiple roles
                if fact_type in fact_type_to_complex_ucs:
                    for uc in fact_type_to_complex_ucs[fact_type]:
                        uc_roles = []
                        for role in fact_type_to_roles[fact_type]:
                            p = md.lookup_concept(object_types[role.player].name)
                            role_idx = role_idx_to_player.index(p)
                            rl = rel.role_at(role_idx)
                            if role.id in uc.roles:
                                uc_roles.append(rl)
                        md.unique(*uc_roles)

    def _add_external_identifying_relationships(self, model):
        roles = self._parser.roles()
        object_types = self._parser.object_types()
        for uc_id, uc in self._parser.external_uniqueness_constraints().items():
            et = model.lookup_concept(object_types[uc.identifies].name)
            identifying_rel = []
            for ro in uc.roles:
                rel_name = roles[ro].relationship_name
                relationship = model.lookup_relationship(rel_name)
                for rl in relationship.roles():
                    if rl.player_type == et:
                        idx = relationship.roles().index(rl)
                        identifying_rel.append(relationship.relations()[idx])
            model.ref_scheme(*identifying_rel)

    def _add_role_value_constraints(self, model):
        for rel, rvc in self._relationship_role_value_constraints.items():
            for rls in rel.relations():
                if isinstance(rls.reading().role_at(0).player(), AbstractEntityType):
                    model.role_value_constraint(rls, values=rvc.values)
                    break

    @staticmethod
    def _map_datatype(dt):
        mapping = {
            "AutoCounterNumericDataType": Integer,
            "UnsignedIntegerNumericDataType": UnsignedInteger,
            "VariableLengthTextDataType": String,
            "DateAndTimeTemporalDataType": DateTime,
            "DecimalNumericDataType": Decimal,
            "DateTemporalDataType": Date
        }
        return mapping.get(dt, String)


"""
    Intermediate Representation of RelationalAI programs.
"""
from __future__ import annotations
from dataclasses import dataclass, field
from enum import Enum
from io import StringIO
from itertools import count

from pandas import DataFrame
from more_itertools import peekable
from typing import Any, Iterator, Optional, Tuple, TypeVar, Union as PyUnion, IO, Iterable
from copy import copy
from itertools import islice
from datetime import datetime, date
from decimal import Decimal as PyDecimal

from .util import FrozenOrderedSet, Printer as BasePrinter, NameCache

import json

#--------------------------------------------------
# IR Nodes
#--------------------------------------------------

_global_id = peekable(count(0))
def next_id():
    return next(_global_id)

def reconstructor(cls):
    """
    A decorator that adds a `reconstruct` method to the class.
    The method reconstructs the object with new values for its fields.
    The `reconstruct` method can take either positional arguments in field order
    (excluding the `id` field) or keyword arguments for specific fields.
    For positional arguments, fields of superclasses precede fields of subclasses.
    If there is no difference between the current values of the fields and the new values,
    `self` is returned, otherwise the object is copied and modified.
    Note that the `id` field is not updated, unlike when creating a new node.
    """
    # Get the names of all fields from the class, except "id".
    fnames = [fname for fname in cls.__dataclass_fields__ if fname != "id"]

    # Add reconstruct method that can handle both positional and keyword arguments
    def reconstruct(self, *args, **kwargs):
        """
        Reconstruct the object with new values for its fields.

        Can be called in two ways:
        1. With positional arguments in the same order as the fields in the class
        2. With keyword arguments for specific fields to update

        It is an error to provide both positional and keyword arguments.
        Returns self if no changes are needed, otherwise returns a new instance.
        """
        # Validate arguments
        if args and kwargs:
            raise ValueError("Cannot provide both positional and keyword arguments")

        # Handle positional arguments
        if args:
            if len(args) != len(fnames):
                raise ValueError(f"Expected {len(fnames)} positional arguments, got {len(args)}")

            # Check if any values differ from current values
            has_changes = False
            for fname, value in zip(fnames, args):
                if getattr(self, fname) is not value:
                    has_changes = True
                    break

            if not has_changes:
                return self

            # Create a new instance with updated values
            new_obj = copy(self)
            for fname, value in zip(fnames, args):
                object.__setattr__(new_obj, fname, value)

            return new_obj

        # Handle keyword arguments
        else:
            # Check if any values differ from current values
            has_changes = False
            for fname, value in kwargs.items():
                if fname not in fnames:
                    raise ValueError(f"Field '{fname}' is not a valid field for this object")
                if getattr(self, fname) is not value:
                    has_changes = True
                    break

            if not has_changes:
                return self

            # Create a new instance with updated values
            new_obj = copy(self)
            for fname, value in kwargs.items():
                if fname in fnames:
                    object.__setattr__(new_obj, fname, value)

            return new_obj

    setattr(cls, "reconstruct", reconstruct)

    # Also add a clone method that will recreate the node with the same data but with a
    # new, distinct node id.
    def clone(self):
        """
        Reconstruct a node with the same values, but it will get a new id.
        """
        args = []
        for fname in fnames:
            args.append(getattr(self, fname))
        return self.__class__(*args)

    setattr(cls, "clone", clone)

    return cls

def acceptor(cls):
    """
    A decorator that adds an `accept` method to the class.
    The `accept` method calls the appropriate handler method
    of a Dispatcher.

    Using a decorator here allows us to compute the handler name once
    for the class rather than each time we dispatch.
    """
    attr_name = f"visit_{cls.__name__.lower()}"
    def accept(self, v, parent=None):
        return getattr(v, attr_name)(self, parent)
    cls.accept = accept
    return cls

def annotations_field() -> FrozenOrderedSet[Annotation]:
    return field(default_factory=lambda: FrozenOrderedSet([]), compare=False, hash=False)

@dataclass(frozen=True)
class Node:
    # A generated id that is not used on comparisons and hashes
    id: int = field(default_factory=next_id, init=False, compare=False, hash=False)

    @property
    def kind(self):
        return self.__class__.__name__.lower()

    def __str__(self):
        return node_to_string(self)

    def accept(self, v, parent=None):
        raise NotImplementedError(f"accept not implemented for {self.kind}")

    def reconstruct(self, *args, **kwargs):
        raise NotImplementedError(f"reconstruct not implemented for {self.kind}")

    def clone(self, *args):
        raise NotImplementedError(f"clone not implemented for {self.kind}")


#-------------------------------------------------
# Public Types - Model
#-------------------------------------------------

@acceptor
@reconstructor
@dataclass(frozen=True)
class Model(Node):
    """Represents the whole universe of elements that make up a program."""
    engines: FrozenOrderedSet["Engine"]
    relations: FrozenOrderedSet["Relation"]
    types: FrozenOrderedSet["Type"]
    root: Task
    annotations:FrozenOrderedSet[Annotation] = annotations_field()


#-------------------------------------------------
# Public Types - Engine
#-------------------------------------------------

@acceptor
@reconstructor
@dataclass(frozen=True)
class Capability(Node):
    """Engine capabilities, such as 'graph algorithms', 'solver', 'constant time count', etc"""
    name: str

@acceptor
@reconstructor
@dataclass(frozen=True)
class Engine(Node):
    """The entity that owns a Task and provides access to certain relations."""
    name: str
    platform: str # SQL, Rel, JS, OpenAI, etc
    info: Any
    capabilities: FrozenOrderedSet[Capability]
    relations: FrozenOrderedSet["Relation"]
    annotations:FrozenOrderedSet[Annotation] = annotations_field()


#-------------------------------------------------
# Public Types - Data Model
#-------------------------------------------------

@acceptor
@reconstructor
@dataclass(frozen=True)
class Type(Node):
    """The type of a field in a relation."""
    pass

@acceptor
@dataclass(frozen=True)
class ScalarType(Type):
    """The named type."""
    name: str
    super_types: FrozenOrderedSet[ScalarType] = field(default_factory=lambda: FrozenOrderedSet([]))
    annotations:FrozenOrderedSet[Annotation] = annotations_field()

    def __eq__(self, other):
        return isinstance(other, ScalarType) and other.id == self.id

    def __hash__(self):
        return hash(self.id)


@acceptor
@reconstructor
@dataclass(frozen=True)
class ListType(Type):
    """A type that represents a list of elements of some other type."""
    element_type: Type

@acceptor
@reconstructor
@dataclass(frozen=True)
class UnionType(Type):
    """A type that represents either one of a set of types."""
    types: FrozenOrderedSet[Type]

    def __eq__(self, other):
        return isinstance(other, UnionType) and len(self.types) == len(other.types) and self.types.includes(other.types)

@acceptor
@dataclass(frozen=True)
class TupleType(Type):
    """A type that represents a tuple of elements of some other types."""
    types: Tuple[Type, ...]

@acceptor
@reconstructor
@dataclass(frozen=True)
class Field(Node):
    """A named field in a relation."""
    name: str
    type: Type
    input: bool # must be grounded as the relation cannot compute it

    def __eq__(self, other):
        return isinstance(other, Field) and other.id == self.id

    def __hash__(self):
        return hash(self.id)


@acceptor
@reconstructor
@dataclass(frozen=True)
class Relation(Node):
    """A relation represents the schema of a set of tuples."""
    name: str
    fields: Tuple[Field, ...]
    requires: FrozenOrderedSet[Capability]
    annotations:FrozenOrderedSet[Annotation] = annotations_field()
    overloads: FrozenOrderedSet[Relation] = field(default_factory=lambda: FrozenOrderedSet([]))

    def __eq__(self, other):
        return isinstance(other, Relation) and other.id == self.id

    def __hash__(self):
        return hash(self.id)


#-------------------------------------------------
# Public Types - Tasks
#-------------------------------------------------

@acceptor
@reconstructor
@dataclass(frozen=True)
class Task(Node):
    engine: Optional[Engine]

#
# Task composition
#

@acceptor
@reconstructor
@dataclass(frozen=True)
class Logical(Task):
    """Execute sub-tasks up to fix-point."""
    # Executes tasks concurrently. Succeeds if every task succeeds.
    hoisted: Tuple[VarOrDefault, ...]
    body: Tuple[Task, ...]
    annotations:FrozenOrderedSet[Annotation] = annotations_field()

@acceptor
@reconstructor
@dataclass(frozen=True)
class Union(Task):
    """Execute sub-tasks in any order."""
    # Executes tasks concurrently. Succeeds if at least one task succeeds.
    hoisted: Tuple[VarOrDefault, ...]
    tasks: Tuple[Task, ...]
    annotations:FrozenOrderedSet[Annotation] = annotations_field()

@acceptor
@reconstructor
@dataclass(frozen=True)
class Sequence(Task):
    """Execute sub-tasks one at a time, in this order."""
    # Executes tasks in order. Stops when a task fails. Succeeds if all tasks succeed.
    hoisted: Tuple[VarOrDefault, ...]
    tasks: Tuple[Task, ...]
    annotations:FrozenOrderedSet[Annotation] = annotations_field()

@acceptor
@reconstructor
@dataclass(frozen=True)
class Match(Task):
    """Execute sub-tasks in order until the first succeeds."""
    # Executes tasks in order. Stops when a task succeeds. Succeeds if some task succeeds.
    hoisted: Tuple[VarOrDefault, ...]
    tasks: Tuple[Task, ...]
    annotations:FrozenOrderedSet[Annotation] = annotations_field()

@acceptor
@reconstructor
@dataclass(frozen=True)
class Check(Node):
    check: Task
    error: Optional[Task]
    annotations:FrozenOrderedSet[Annotation] = annotations_field()

@acceptor
@reconstructor
@dataclass(frozen=True)
class Require(Task):
    """Execute the domain task and verify the checks. If a check is not satisfied, execute its error task."""
    domain: Task
    checks: Tuple[Check, ...]
    annotations:FrozenOrderedSet[Annotation] = annotations_field()

@acceptor
@reconstructor
@dataclass(frozen=True)
class Until(Task):
    """Execute both `check` and `body` concurrently, until check succeeds."""
    hoisted: Tuple[VarOrDefault, ...]
    check: Task
    body: Task
    annotations:FrozenOrderedSet[Annotation] = annotations_field()

@acceptor
@reconstructor
@dataclass(frozen=True)
class Wait(Task):
    hoisted: Tuple[VarOrDefault, ...]
    check: Task
    annotations:FrozenOrderedSet[Annotation] = annotations_field()

# TODO: DynamicLookup


#
# Logical Quantifiers
#

@acceptor
@reconstructor
@dataclass(frozen=True)
class Not(Task):
    """Logical negation of the sub-task."""
    task: Task
    annotations:FrozenOrderedSet[Annotation] = annotations_field()

@acceptor
@reconstructor
@dataclass(frozen=True)
class Exists(Task):
    """Existential quantification over the sub-task."""
    vars: Tuple[Var, ...]
    task: Task
    annotations:FrozenOrderedSet[Annotation] = annotations_field()

@acceptor
@reconstructor
@dataclass(frozen=True)
class ForAll(Task):
    """Universal quantification over the sub-task."""
    vars: Tuple[Var, ...]
    task: Task
    annotations:FrozenOrderedSet[Annotation] = annotations_field()

#
# Iteration (Loops)
#

# loops body until a break condition is met
@acceptor
@reconstructor
@dataclass(frozen=True)
class Loop(Task):
    """Execute the body in a loop, incrementing the iter variable, until a break sub-task in
    the body succeeds."""
    hoisted: Tuple[VarOrDefault, ...]
    iter: Var
    body: Task
    annotations:FrozenOrderedSet[Annotation] = annotations_field()

@acceptor
@reconstructor
@dataclass(frozen=True)
class Break(Task):
    """Break a surrounding Loop if the `check` succeeds."""
    check: Task
    annotations:FrozenOrderedSet[Annotation] = annotations_field()


#
# Relational Operations
#

@acceptor
@reconstructor
@dataclass(frozen=True)
class Var(Node):
    """A named variable that can point to objects of this type."""
    type: Type
    name: str

    def __eq__(self, other):
        return isinstance(other, Var) and other.id == self.id

    def __hash__(self):
        return hash(self.id)


@acceptor
@reconstructor
@dataclass(frozen=True)
class Default(Node):
    """A variable with a default value. This can be used in 'hoisted' attributes to
    represent the value to assign a variable if the underlying task fails."""
    var: Var
    value: Value

VarOrDefault = PyUnion[Var, Default]

@acceptor
@reconstructor
@dataclass(frozen=True)
class Literal(Node):
    """A literal value with a specific type."""
    type: Type
    value: Any

@acceptor
@reconstructor
@dataclass(frozen=True)
class Data(Task):
    """A table of data"""
    data: DataFrame
    vars: tuple[Var, ...]

    def __eq__(self, other):
        return (
            isinstance(other, Data)
            and self.data is other.data
            and self.vars == other.vars
        )

    def __hash__(self):
        # Use id() to get the reference of the DataFrame
        data_ref_hash = hash(id(self.data))
        vars_hash = hash(self.vars)
        return data_ref_hash ^ vars_hash

    def __len__(self):
        return len(self.data)

    def __iter__(self) -> Iterator[tuple[Any, ...]]:
        return iter(self.data.itertuples(index=True))

PyValue = PyUnion[str, int, float, PyDecimal, bool, date, datetime, None]
Value = PyUnion[PyValue, Var, Default, Literal, Type, Relation, Tuple["Value", ...], FrozenOrderedSet["Value"]]

@acceptor
@reconstructor
@dataclass(frozen=True)
class Annotation(Node):
    """Meta information that can be attached to Updates."""
    relation: Relation
    args: Tuple[Value, ...]

class Effect(Enum):
    """Possible effects of an Update."""
    derive = "derive"
    insert = "insert"
    delete = "delete"

@acceptor
@reconstructor
@dataclass(frozen=True)
class Update(Task):
    """Updates the relation with these arguments. The update can derive new tuples
    temporarily, can insert new tuples persistently, or delete previously persisted tuples."""
    relation: Relation
    args: Tuple[Value, ...]
    effect: Effect
    annotations:FrozenOrderedSet[Annotation] = annotations_field()

@acceptor
@reconstructor
@dataclass(frozen=True)
class Lookup(Task):
    """Lookup tuples from this relation, filtering with these arguments."""
    relation: Relation
    args: Tuple[Value, ...]
    annotations:FrozenOrderedSet[Annotation] = annotations_field()

@acceptor
@reconstructor
@dataclass(frozen=True)
class Output(Task):
    """Output the value of these vars, giving them these column names."""
    aliases: FrozenOrderedSet[Tuple[str, Value]]
    keys: Optional[FrozenOrderedSet[Var]]
    annotations:FrozenOrderedSet[Annotation] = annotations_field()

@acceptor
@reconstructor
@dataclass(frozen=True)
class Construct(Task):
    """Construct an id from these values, and bind the id to this var."""
    values: Tuple[Value, ...]
    id_var: Var
    annotations:FrozenOrderedSet[Annotation] = annotations_field()

@acceptor
@reconstructor
@dataclass(frozen=True)
class Aggregate(Task):
    """Perform an aggregation with these arguments."""
    aggregation: Relation
    projection: Tuple[Var, ...]
    group: Tuple[Var, ...]
    args: Tuple[Value, ...]
    annotations:FrozenOrderedSet[Annotation] = annotations_field()

@acceptor
@reconstructor
@dataclass(frozen=True)
class Rank(Task):
    """Perform a rank with these arguments."""
    projection: Tuple[Var, ...]
    group: Tuple[Var, ...]
    args: Tuple[Var, ...]
    # True if the argument is ascending, False if descending
    arg_is_ascending: Tuple[bool, ...]
    result: Var
    limit: int = 0
    annotations:FrozenOrderedSet[Annotation] = annotations_field()

#--------------------------------------------------
# Printer
#--------------------------------------------------

infix = ["+", "-", "*", "/", "%", "=", "!=", "<", "<=", ">", ">="]

T = TypeVar('T', bound=Node)
def node_to_string(node:Node|Tuple[T, ...]|FrozenOrderedSet[T], print_ids=False) -> str:
    io = StringIO()
    printer = Printer(io)
    printer.pprint(node, print_ids=print_ids)
    return io.getvalue()

def value_to_string(value:PyUnion[Value, Tuple[Value, ...]]) -> str:
    return Printer().value_to_string(value)

def type_to_string(t:Type) -> str:
    return value_to_string(t).strip()

def types_to_string(ts: Iterable[Type]) -> str:
    """Join a collection of types with commas."""
    return ', '.join(type_to_string(t) for t in ts)

@dataclass(frozen=True)
class Printer(BasePrinter):
    name_cache: NameCache = field(default_factory=NameCache)

    def print_hoisted(self, depth, name, hoisted: Tuple[VarOrDefault, ...]):
        if hoisted:
            self._indent_print_nl(depth, f"{name} ⇑[{', '.join([self.value_to_string(h) for h in hoisted])}]")
        else:
            self._indent_print_nl(depth, name)

    def value_to_string(self, value:PyUnion[Value, Tuple[Value, ...]]) -> str:
        """ Return a string representation of the value. """
        if isinstance(value, (int, float, bool)):
            return json.dumps(value)
        if isinstance(value, str):
            return f"\"{value}\""
        if isinstance(value, PyDecimal):
            return str(value)
        if isinstance(value, (date, datetime)):
            return json.dumps(value.isoformat())
        elif value is None:
            return "None"
        elif isinstance(value, Var):
            return self.name_cache.get_name(value.id, value.name)
        elif isinstance(value, Default):
            return f"{self.name_cache.get_name(value.var.id, value.var.name)}={value.value}"
        elif isinstance(value, Literal):
            return self.value_to_string(value.value)
        elif isinstance(value, ListType):
            return f"[{self.value_to_string(value.element_type)}]"
        elif isinstance(value, UnionType):
            return f"{{{'; '.join(map(self.value_to_string, value.types))}}}"
        elif isinstance(value, TupleType):
            return f"({', '.join(map(self.value_to_string, value.types))})"
        elif isinstance(value, ScalarType):
            return f"{value.name}"
        elif isinstance(value, Relation):
            return value.name
        elif isinstance(value, Tuple):
            return f"({', '.join(map(self.value_to_string, value))})"
        elif isinstance(value, FrozenOrderedSet):
            return f"{{{', '.join(map(self.value_to_string, value))}}}"
        elif isinstance(value, DataFrame):
            length = len(value)
            vals = islice(value, min(length, 3))
            vals_str = ", ".join([f"({', '.join(map(self.value_to_string, v))})" for v in vals]) # type: ignore
            if length > 3:
                vals_str += f", ... ({length} total)"
            final_str = f"{{{vals_str}}}"
            return final_str
        else:
            raise NotImplementedError(f"value_to_string not implemented for {type(value)}")

    T = TypeVar('T', bound=Node)
    def pprint(self, node:Node|Tuple[T, ...]|FrozenOrderedSet[T], depth=0, print_ids=False) -> None:
        """ Pretty print the node into the io, starting with indentation based on depth. If io is None,
        print into the standard output. """

        if print_ids and not (isinstance(node, Tuple) or isinstance(node, FrozenOrderedSet)):
            self._print(f"|{node.id}|")

        if isinstance(node, Tuple) or isinstance(node, FrozenOrderedSet):
            for n in node:
                self.pprint(n, depth, print_ids=print_ids)
        # model
        elif isinstance(node, Model):
            self._indent_print_nl(depth, "Model")
            if len(node.engines) > 0:
                self._indent_print_nl(depth + 1, "engines:")
            self.pprint(node.engines, depth + 2, print_ids=print_ids)
            if len(node.relations) > 0:
                self._indent_print_nl(depth + 1, "relations:")
            self.pprint(node.relations, depth + 2, print_ids=print_ids)
            if len(node.types) > 0:
                self._indent_print_nl(depth + 1, "types:")
            self.pprint(node.types, depth + 2, print_ids=print_ids)
            self._indent_print_nl(depth + 1, "root:")
            self.pprint(node.root, depth + 2, print_ids=print_ids)

        # engine
        elif isinstance(node, Capability):
            self._indent_print_nl(depth, node.name)
        elif isinstance(node, Engine):
            self._indent_print_nl(depth, f"Engine ({node.name}, {node.platform})")
            self._indent_print_nl(depth + 1, node.info)
            self._indent_print_nl(depth + 1, ', '.join([c.name for c in node.capabilities]))
            self.pprint(node.relations, depth + 1, print_ids=print_ids)

        # data model
        elif isinstance(node, Type):
            self._indent_print_nl(depth, self.value_to_string(node))
        elif isinstance(node, Field):
            s = f"{node.name}: {self.value_to_string(node.type)}{' (input)' if node.input else ''}"
            self._indent_print_nl(depth, s)
        elif isinstance(node, Relation):
            self._indent_print_nl(depth, node.name)
            for anno in node.annotations:
                self._indent_print(depth + 1, anno)
            self.pprint(node.fields, depth + 1, print_ids=print_ids)
            if len(node.requires) > 0:
                self._indent_print_nl(depth + 1, "requires:")
                self.pprint(node.requires, depth + 2, print_ids=print_ids)

        # tasks

        # Task composition
        elif isinstance(node, Logical):
            self.print_hoisted(depth, "Logical", node.hoisted)
            self.pprint(node.body, depth + 1, print_ids=print_ids)
        elif isinstance(node, Sequence):
            self.print_hoisted(depth, "Sequence", node.hoisted)
            self.pprint(node.tasks, depth + 1, print_ids=print_ids)
        elif isinstance(node, Union):
            self.print_hoisted(depth, "Union", node.hoisted)
            self.pprint(node.tasks, depth + 1, print_ids=print_ids)
        elif isinstance(node, Match):
            self.print_hoisted(depth, "Match", node.hoisted)
            self.pprint(node.tasks, depth + 1, print_ids=print_ids)
        elif isinstance(node, Until):
            self.print_hoisted(depth, "Until", node.hoisted)
            self.pprint(node.check, depth + 1, print_ids=print_ids)
            self.pprint(node.body, depth + 1, print_ids=print_ids)
        elif isinstance(node, Wait):
            self.print_hoisted(depth, "Match", node.hoisted)
            self.pprint(node.check, depth + 1, print_ids=print_ids)
        elif isinstance(node, Require):
            self._indent_print_nl(depth, "Require")
            self.pprint(node.domain, depth + 1, print_ids=print_ids)
            for check in node.checks:
                self.pprint(check, depth + 1, print_ids=print_ids)
        elif isinstance(node, Check):
            self._indent_print_nl(depth, "Check")
            self.pprint(node.check, depth + 1, print_ids=print_ids)
            if node.error:
                self.pprint(node.error, depth + 1, print_ids=print_ids)
            else:
                self._indent_print_nl(depth + 1, "None")

        # Relational Operations
        elif isinstance(node, Var):
            self._indent_print_nl(0, self.value_to_string(node))
        elif isinstance(node, Default):
            self._indent_print_nl(depth, f"{self.value_to_string(node.var)} (default: {self.value_to_string(node.value)})")
        elif isinstance(node, Literal):
            self._indent_print_nl(0, self.value_to_string(node))
        elif isinstance(node, Annotation):
            if node.args:
                self._indent_print_nl(depth, f"@{node.relation.name}{self.value_to_string(node.args)}")
            else:
                self._indent_print_nl(depth, f"@{node.relation.name}")
        elif isinstance(node, Update):
            rel_name = node.relation.name
            annos = "" if not node.annotations else f" {' '.join(f'@{a.relation.name}' for a in node.annotations)}"
            self._indent_print_nl(depth, f"→ {node.effect.value} {rel_name}{self.value_to_string(node.args)}{annos}")
        elif isinstance(node, Lookup):
            rel_name = node.relation.name
            annos = "" if not node.annotations else f" {' '.join(f'@{a.relation.name}' for a in node.annotations)}"
            if rel_name in infix:
                args = [self.value_to_string(arg) for arg in node.args]
                if len(node.args) == 2:
                    self._indent_print_nl(depth, f"{args[0]} {rel_name} {args[1]}{annos}")
                elif len(node.args) == 1:
                    self._indent_print_nl(depth, f"{rel_name}{args[0]}{annos}")
                elif len(node.args) == 3:
                    self._indent_print_nl(depth, f"{args[2]} = {args[0]} {rel_name} {args[1]}{annos}")
            else:
                self._indent_print_nl(depth, f"{rel_name}{self.value_to_string(node.args)}{annos}")
        elif isinstance(node, Output):
            keys = []
            if node.keys:
                for k in node.keys:
                    keys.append(self.value_to_string(k))
            args = []
            for k, v in node.aliases:
                ppv = self.value_to_string(v)
                if k != ppv:
                    args.append(f"{ppv} as '{k}'")
                else:
                    args.append(ppv)
            keys_str = f"[{', '.join(keys)}]" if keys else ""
            annos = "" if not node.annotations else f" {' '.join(f'@{a.relation.name}' for a in node.annotations)}"
            self._indent_print_nl(depth, f"→ output{keys_str}({', '.join(args)}){annos}")
        elif isinstance(node, Construct):
            values = [self.value_to_string(v) for v in node.values]
            self._indent_print_nl(depth, f"construct({', '.join(values)}, {self.value_to_string(node.id_var)})")
        elif isinstance(node, Aggregate):
            projection = ", ".join([self.value_to_string(v) for v in node.projection])
            group = ", ".join([self.value_to_string(v) for v in node.group])
            args = ", ".join([self.value_to_string(v) for v in node.args])
            self._indent_print_nl(depth, f"{node.aggregation.name}([{projection}], [{group}], [{args}])")
        elif isinstance(node, Rank):
            def str_rank_var(v, o):
                up, down = "'↑'", "'↓'"
                return f"{self.value_to_string(v)}{up if o else down}"
            projection = ", ".join([self.value_to_string(v) for v in node.projection])
            group = ", ".join([self.value_to_string(v) for v in node.group])
            assert len(node.args) == len(node.arg_is_ascending)
            args = ", ".join([str_rank_var(v, o) for v, o in zip(node.args, node.arg_is_ascending)])
            result = self.value_to_string(node.result)
            limit = f"[limit={node.limit}]" if node.limit != 0 else ""
            self._indent_print_nl(depth, f"rank{limit}([{projection}], [{group}], [{args}], {result})")
        elif isinstance(node, Data):
            length = len(node)
            vals = islice(node, min(length, 3))
            vals_str = ", ".join([f"({', '.join(map(self.value_to_string, v))})" for v in vals])
            if length > 3:
                vals_str += f", ... ({length} total)"
            vars_str = ", ".join([self.value_to_string(v) for v in node.vars])
            final_str = f"{{{vals_str}}}({vars_str})"
            self._indent_print_nl(depth, final_str)

        # Logical Quantifiers
        elif isinstance(node, Not):
            self._indent_print_nl(depth, "Not")
            self.pprint(node.task, depth + 1, print_ids=print_ids)
        elif isinstance(node, Exists):
            self._indent_print_nl(depth, f"Exists({', '.join([self.value_to_string(v) for v in node.vars])})")
            self.pprint(node.task, depth + 1, print_ids=print_ids)
        elif isinstance(node, ForAll):
            self._indent_print_nl(depth, f"ForAll({', '.join([self.value_to_string(v) for v in node.vars])})")
            self.pprint(node.task, depth + 1, print_ids=print_ids)

        # Iteration (Loops)
        elif isinstance(node, Loop):
            self.print_hoisted(depth, f"Loop ⇓[{self.value_to_string(node.iter)}]", node.hoisted)
            self.pprint(node.body, depth + 1, print_ids=print_ids)

        elif isinstance(node, Break):
            self._indent_print_nl(depth, "Break")
            self.pprint(node.check, depth + 1, print_ids=print_ids)

        elif isinstance(node, Task):
            self._indent_print_nl(depth, "Success")

        else:
            # return
            raise NotImplementedError(f"pprint not implemented for {type(node)}")

def dump_to_string(node: Node|Tuple[T, ...]|FrozenOrderedSet[T], depth=0, io: Optional[IO[str]] = None) -> str:
    """Dump a node to a string."""
    output = StringIO()
    dump(node, depth, output)
    return output.getvalue()


def dump(node: Node|Tuple[T, ...]|FrozenOrderedSet[T], depth=0, io: Optional[IO[str]] = None) -> None:
    """Print complete IR information, including node IDs and all fields."""

    def indent_print(depth, io: Optional[IO[str]], *args) -> None:
        """ Helper to print the arguments into the io with indented based on depth. """
        if io is None:
            print("    " * depth + " ".join(map(str, args)))
        else:
            io.write("    " * depth + " ".join(map(str, args)) + "\n")

    def print_node_header(n: Node, depth: int):
        indent_print(depth, io, f"{n.__class__.__name__} (id={n.id})")

    if isinstance(node, (Tuple, FrozenOrderedSet)):
        for n in node:
            if isinstance(n, Tuple) and len(n) == 2 and isinstance(n[0], str) and isinstance(n[1], Node):
                indent_print(depth, io, f"{n[0]}:")
                dump(n[1], depth + 1, io)
            elif isinstance(n, (Node, Tuple, FrozenOrderedSet)):
                dump(n, depth, io)
            else:
                indent_print(depth, io, f"{value_to_string(n)}")
        return

    if not isinstance(node, Node):
        raise TypeError(f"Expected Node but got {type(node)}")

    print_node_header(node, depth)

    # Print all fields of the dataclass
    for field_name, field_value in node.__dict__.items():
        if field_name == 'id':  # Already printed in header
            continue

        assert not isinstance(field_value, list), f"{node.__class__.__name__}.{field_name} is a {type(field_value)}"
        assert not isinstance(field_value, set), f"{node.__class__.__name__}.{field_name} is a {type(field_value)}"

        if isinstance(field_value, (Node, Tuple, FrozenOrderedSet)) and not isinstance(field_value, Type):
            indent_print(depth + 1, io, f"{field_name}:")
            dump(field_value, depth + 2, io)
        elif isinstance(field_value, Effect):
            indent_print(depth + 1, io, f"{field_name}: {field_value.value}")
        else:
            indent_print(depth + 1, io, f"{field_name}: {value_to_string(field_value)}")



from __future__ import annotations
import base64
from collections import defaultdict
from dataclasses import dataclass, field
from enum import Enum
import numbers
from typing import Dict, Iterable, List, Optional, Set, Tuple, Union, Any, cast
import json
import textwrap
from datetime import datetime, date
import rich

#--------------------------------------------------
# Base
#--------------------------------------------------

id = 0
def next_id():
    global id
    id += 1
    return id

@dataclass
class Base():
    id: int = field(default_factory=next_id, init=False)

    def __str__(self):
        return Printer().print(self)

#--------------------------------------------------
# Data
#--------------------------------------------------

@dataclass
class Type(Base):
    name: str = ""
    properties: List['Property'] = field(default_factory=list)
    parents: List['Type'] = field(default_factory=list)
    agent: Optional['Agent'] = None

    def isa(self, other: 'Type') -> bool:
        if self == other:
            return True
        for p in self.parents:
            if p.isa(other):
                return True
        return False

    def __hash__(self) -> int:
        return self.id

UnknownType = Type("Unknown")

@dataclass
class Property(Base):
    name: str
    type: Type
    is_input: bool = False
    parents: List[Type] = field(default_factory=list)

    def __hash__(self) -> int:
        return self.id

    @staticmethod
    def find(name:str, types:List[Type], ignore_case=False) -> Optional['Property']:
        found = None
        for t in types:
            if found:
                break
            for p in t.properties:
                if p.name == name or (ignore_case and p.name.lower() == name.lower()):
                    found = p
                    break
        return found

@dataclass
class Agent(Base):
    name: str
    platform: str  # SQL, Rel, JS, OpenAI, etc
    info: Any

    def __hash__(self) -> int:
        return self.id

#--------------------------------------------------
# Task
#--------------------------------------------------

class Behavior(Enum):
    Query = "query"
    Union = "union"
    OrderedChoice = "ordered_choice"
    Sequence = "sequence"
    Catch = "catch"

@dataclass
class Task(Type):
    behavior: Behavior = Behavior.Query
    items: List['Action'] = field(default_factory=list)
    bindings: Dict[Property, 'Var'] = field(default_factory=dict)
    inline: bool = False

    _task_builtin: Any = None
    def __post_init__(self):
        self.parents.append(Task._task_builtin)

    def __hash__(self) -> int:
        return self.id

    def return_cols(self, allow_dups=True) -> List[str]:
        namer = Namer()
        for i in self.items:
            ent = i.entity.value
            if i.action == ActionType.Call and isinstance(ent, Task):
                sub = ent.return_cols()
                if len(sub):
                    return sub
            if i.action == ActionType.Bind and i.entity.isa(Builtins.Return):
                cols = []
                for (prop, var) in i.bindings.items():
                    if allow_dups:
                        cols.append(namer.get(var) or namer.get(prop))
                    else:
                        cols.append(namer.get_safe_name(var.name or prop.name))
                return cols
        return []

    def has_persist(self, seen=None) -> bool:
        if seen is None:
            seen = set()
        for i in self.items:
            if i.action in [ActionType.Persist, ActionType.Unpersist]:
                return True
            vs = i.vars()
            for v in vs:
                if v not in seen:
                    seen.add(v)
                    if v.value and isinstance(v.value, Task) and v.value.has_persist(seen):
                        return True
        return False

    def requires_provides(self, seen = None) -> Tuple[Set['Var'], Set['Var'], Set['Var']]:
        sub_requires = set()
        sub_provides = set()
        refs = set()
        for i in self.items:
            r, p, sub_refs = i.requires_provides(seen)
            sub_requires.update(r)
            sub_provides.update(p)
            refs.update(p | sub_refs)
        return sub_requires - sub_provides, sub_provides, refs

    def normalize(self):
        entity_action:Dict[Any, Action] = {}
        seen = set()
        collapsed:List[Action] = []
        for i in self.items:
            if i in seen:
                continue # prevent duplicates
            seen.add(i)

            # recurse on subtasks
            ent = i.entity.value
            if isinstance(ent, Task):
                ent.normalize()
            for (_, var) in i.bindings.items():
                if isinstance(var.value, Task):
                    var.value.normalize()

            # we can only collapse get, bind, persist, unpersist
            if i.action not in [ActionType.Get, ActionType.Bind, ActionType.Persist, ActionType.Unpersist]:
                collapsed.append(i)
                continue

            found = entity_action.get((i.entity, i.action))
            if not found:
                if len(i.types) or len(i.bindings):
                    entity_action[(i.entity, i.action)] = i
                    collapsed.append(i)
            else:
                # merge with existing
                exists = False
                for key, val in i.bindings.items():
                    found_val = found.bindings.get(key)
                    if found_val is not None and found_val is not val: #and (not found_val.value or found_val.value is not val.value):
                        exists = True
                if not exists:
                    found.types.extend([x for x in i.types if x not in found.types])
                    found.bindings.update(i.bindings)
                else:
                    collapsed.append(i)

        if self.behavior == Behavior.Query:
            try:
                self.items = Utils.sort(collapsed)
            except Exception:
                rich.print("[red bold]Stratification failed[/red bold][red]\nThings may not work correctly. Please send us this program so we can take a look!\n")
                self.items = collapsed
        else:
            self.items = collapsed

#--------------------------------------------------
# Var
#--------------------------------------------------

Value = Union[str, numbers.Number, int, float, bool, Task, Property, Type, bytes,
              datetime, date, List['Value'], List['Var']]

@dataclass
class Var(Base):
    id: int = field(default_factory=next_id, init=False)
    type: Type = UnknownType
    name: Optional[str] = None
    value: Optional[Value] = None

    def __hash__(self) -> int:
        return self.id

    def __eq__(self, other: object) -> bool:
        if type(other) is not Var:
            return False
        if super().__eq__(other):
            return True
        if other.value is not None and self.value is not None:
            return other.value is self.value
        return False

    def isa(self, other: Type) -> bool:
        return bool(self.type.isa(other) or (self.value and isinstance(self.value, Type) and self.value.isa(other)))

#--------------------------------------------------
# Action
#--------------------------------------------------

class ActionType(Enum):
    Get = "get"
    Call = "call"
    Persist = "persist"
    Unpersist = "unpersist"
    Bind = "bind"
    Construct = "construct"

    def is_effect(self):
        return self in [ActionType.Bind, ActionType.Persist, ActionType.Unpersist]

@dataclass
class Action(Base):
    action: ActionType
    entity: Var
    types: List[Type] = field(default_factory=list)
    bindings: Dict[Property, Var] = field(default_factory=dict)

    def __post_init__(self):
        from . import debugging
        debugging.set_source(self)

    def requires_provides(self, seen = None) -> Tuple[Set[Var], Set[Var], Set[Var]]:
        requires = set()
        provides = set()
        refs = set()
        if not seen:
            seen = set()
        if self in seen:
            return requires, provides, refs

        seen.add(self)
        is_sub_task = self.is_subtask_call()
        if is_sub_task:
            sub_requires = set()
            sub_provides = set()
            for i in cast(Task, self.entity.value).items:
                r, p, sub_refs = i.requires_provides(seen)
                sub_requires.update(r)
                sub_provides.update(p)
                refs.update(p | sub_refs)
            requires.update(sub_requires - sub_provides)

        if self.action == ActionType.Get and self.entity.value is None:
            if len(self.types):
                provides.add(self.entity)
            else:
                requires.add(self.entity)
        elif self.action != ActionType.Call and self.entity.value is None:
            requires.add(self.entity)

        items = [*self.bindings.items()]
        if self.action == ActionType.Construct:
            provides.add(items[-1][1])
            items = items[:-1]

        for k,var in items:
            into = provides
            if self.action in [ActionType.Bind, ActionType.Persist, ActionType.Unpersist, ActionType.Construct]:
                into = requires
            elif k.is_input:
                into = requires

            if var.value is not None and isinstance(var.value, list):
                for v in var.value:
                    if isinstance(v, Var) and v.value is None:
                        into.add(v)
            elif var.value is None:
                into.add(var)

        refs.update(requires)
        return requires - provides, provides, refs

    def vars(self, recursive=False, with_subs=False) -> Set[Var]:
        vars = set(self.bindings.values())
        vars.add(self.entity)
        if recursive:
            for v in list(vars):
                if isinstance(v.value, list):
                    for var in v.value:
                        if isinstance(var, Var) and var.value is None:
                            vars.add(var)
        if with_subs and self.is_subtask_call():
            for i in cast(Task, self.entity.value).items:
                vars.update(i.vars(recursive, with_subs))
        return vars

    def is_subtask_call(self) -> bool:
        return self.action == ActionType.Call \
                and isinstance(self.entity.value, Task) \
                and len(self.entity.value.items) > 0

    def append(self, item: Union[Type, Property, Task, Agent, Var], var: Optional[Var] = None):
        if isinstance(item, Type):
            self.types.append(item)
        elif isinstance(item, Property) and var:
            self.bindings[item] = var
        elif isinstance(item, Var):
            self.entity = item

    def equiv(self, other: 'Action') -> bool:
        if self is other:
            return True
        return self.action == other.action \
                and self.entity == other.entity \
                and self.types == other.types \
                and self.bindings == other.bindings

    def params_list(self) -> List[Var]:
        return list(self.bindings.values())

    def __hash__(self) -> int:
        return self.id

#--------------------------------------------------
# All
#--------------------------------------------------

AllItems = Union[Type, Property, Task, Agent, Var, Action]

#--------------------------------------------------
# Builtins
#--------------------------------------------------

class BuiltinsClass:
    def __init__(self) -> None:
        self.Primitive = Type("Primitive")

        self.Unknown = UnknownType
        self.Any = Type("Any")
        self.String = Type("String", parents=[self.Primitive])
        self.Number = Type("Number", parents=[self.Primitive])
        self.Int = Type("Int", parents=[self.Number])
        self.Decimal = Type("Decimal", parents=[self.Number])
        self.Bool = Type("Bool", parents=[self.Primitive])
        self.Type = Type("Type", parents=[self.Primitive])
        self.Property = Type("Property", parents=[self.Primitive])
        self.ValueType = Type("ValueType", parents=[self.Type])
        self.Symbol = Type("Type", parents=[self.Primitive])
        self.Missing = Type("Missing", parents=[self.Primitive])
        self.Relation = Type("Relation", parents=[self.Primitive], properties=[Property(f"v{i}", self.Any) for i in range(1000)])
        self.Anonymous = Type("Anonymous") # A thing we assume to exist in the host DB for which we don't have information.
        self.Task = Type("Task", parents=[self.Type, self.Relation])
        Task._task_builtin = self.Task
        self.Install = Task("Install", parents=[self.Task], properties=[Property("item", self.Any)])
        self.Distinct = Type("Distinct", parents=[])
        self.Return = Type("Return", parents=[self.Relation], properties=[Property(f"v{i}", self.Any) for i in range(20)])
        self.ReturnDistinct = Type("ReturnDistinct", parents=[self.Return, self.Distinct], properties=[Property(f"v{i}", self.Any) for i in range(20)])
        self.ExportReturn = Type("ExportReturn", parents=[self.Return], properties=[Property(f"v{i}", self.Any) for i in range(20)])
        self.ExportReturnDistinct = Type("ExportReturnDistinct", parents=[self.ExportReturn, self.Distinct], properties=[Property(f"v{i}", self.Any) for i in range(20)])
        self.RawCode = Type("RawCode", properties=[Property("code", self.String)])
        self.RawData = Type("RawData", parents=[self.Relation])
        self.Inline = Type("Inline", parents=[])
        self.Intermediate = Type("Intermediate", parents=[self.Relation])
        self.InlineRawData = Type("InlineRawData", parents=[self.RawData])
        self.InlineExpression = Type("InlineExpression", parents=[])
        self.Aggregate = Type("Aggregate", parents=[self.Task])
        self.Extender = Type("Extender", parents=[self.Aggregate])

        self.Annotation = Type("Annotation")
        self.InspectOptimizedSimple = Type("Inspect(:rel, 1)", parents=[self.Annotation])
        self.InspectOptimizedVerbose = Type("Inspect(:rel, 2)", parents=[self.Annotation])
        self.InlineAnnotation = Type("Inline", parents=[self.Annotation])
        self.NoInlineAnnotation = Type("No_Inline", parents=[self.Annotation])
        self.PipelineAnnotation = Type("Pipeline", parents=[self.Annotation])
        self.FunctionAnnotation = Type("Function", parents=[self.Annotation])
        self.PQEdgeAnnotation = Type("PQ_Edge", parents=[self.Annotation])
        self.PQFilterAnnotation = Type("PQ_Filter", parents=[self.Annotation])
        self.TrackConnAnnotation = Type("Track(:graphlib, :pathfinder_conn)", parents=[self.Annotation])

        self.Expensive = Type("Expensive")

        self.SingleValued = Type("SingleValued")
        self.MultiValued = Type("MultiValued")

        self.Identity = Type("Identity", parents=[self.String])
        self.make_identity = Task("make_identity", properties=[
            Property("params", self.Any, True),
            Property("identity", self.Any)
        ], parents=[self.Expensive])

        self.Infix = Type("Infix")
        self.Filter = Type("Filter")
        self.GlobalFilter = Type("GlobalFilter")

        self.Quantifier = Type("Quantifier", parents=[self.Task, self.GlobalFilter])
        self.Not = Task("Not", parents=[self.Quantifier], properties=[Property("group", self.Any, True), Property("task", self.Task, True)])
        self.Exists = Task("Exists", parents=[self.Quantifier], properties=[Property("group", self.Any, True), Property("task", self.Task, True)])
        self.Every = Task("Every", parents=[self.Quantifier], properties=[Property("group", self.Any, True), Property("task", self.Task, True)])

        self.NonFiltering = Type("NonFiltering", parents=[])
        self.BoolFilter = Type("pyrel_bool_filter", parents=[self.Relation, self.NonFiltering, self.SingleValued], properties=[
            Property("left", self.Any, True),
            Property("right", self.Any, True),
            Property("result", self.Bool, False),
            Property("relation", self.Relation, True),
        ])
        self.Default = Type("pyrel_default", parents=[self.Relation, self.NonFiltering, self.SingleValued], properties=[
            Property("relation", self.Relation, True),
            Property("default", self.Any, True),
            Property("ent", self.Any, True),
            Property("result", self.Any, False),
        ])

        self.ExternalInput = Type("External")
        self.EDB = Type("EDB", parents=[])

        def binary_op(op, with_result=False):
            t = Task(name=op, parents=[self.Infix, self.SingleValued], properties=[
                Property("a", self.Number, True),
                Property("b", self.Number, True),
            ])
            if with_result:
                t.properties.append(Property("result", self.Number))
            else:
                t.parents.append(self.Filter)
            return t

        def aggregate(op):
            return Task(name=op, parents=[self.Aggregate], properties=[
                Property("projection", self.Any, True),
                Property("group", self.Any, True),
                Property("pre_args", self.Any, True),
                Property("result", self.Number),
            ])

        self.gt = binary_op(">")
        self.gte = binary_op(">=")
        self.lt = binary_op("<")
        self.lte = binary_op("<=")
        self.eq = binary_op("=")
        self.approx_eq = binary_op("~=")
        self.neq = binary_op("!=")

        self.plus = binary_op("+", True)
        self.minus = binary_op("-", True)
        self.mult = binary_op("*", True)
        self.div = binary_op("/", True)
        self.floor_div = binary_op("::std::common::floor_divide", True)
        self.floor_div.parents = []

        self.pow = binary_op("^", True)
        self.mod = binary_op("::std::common::modulo", True)
        self.mod.parents = []

        self.count = aggregate("::std::common::count")
        self.prod = aggregate("::std::common::product")
        self.sum = aggregate("::std::common::sum")
        self.avg = aggregate("::std::common::average")
        self.min = aggregate("::std::common::min")
        self.max = aggregate("::std::common::max")

Builtins = BuiltinsClass()

#--------------------------------------------------
# Builder
#--------------------------------------------------

class Builder():
    def __init__(self, to_var):
        self.to_var = to_var

    def call(self, op, params:List[Any]):
        a = Action(ActionType.Call, Var(Builtins.Task, value=op))
        for ix, p in enumerate(params):
            a.append(op.properties[ix], self.to_var(p))
        return a

    def return_(self, params:List[Any], export=False, distinct=False):
        if export and distinct:
            return self.relation_action(ActionType.Bind, Builtins.ExportReturnDistinct, params)
        elif export:
            return self.relation_action(ActionType.Bind, Builtins.ExportReturn, params)
        elif distinct:
            return self.relation_action(ActionType.Bind, Builtins.ReturnDistinct, params)
        else:
            return self.relation_action(ActionType.Bind, Builtins.Return, params)

    def relation(self, name:str, field_count:int):
        return Type(name,
                    parents=[Builtins.Relation, Builtins.Anonymous],
                    properties=[Builtins.Relation.properties[i] for i in range(field_count)])

    def relation_action(self, action_type:ActionType, op:Type|Property, params:Iterable[Any]):
        a = Action(action_type, Var(Builtins.Type, value=op))
        for ix, p in enumerate(params):
            a.append(Builtins.Relation.properties[ix], self.to_var(p))
        return a

    def ident(self, action:Action, omit_type_in_hash=False) -> List[Action]:
        params = []
        if not omit_type_in_hash:
            params = [Var(value=t.name) for t in action.types]
        params.extend(action.bindings.values())
        actions = []
        ident = action.entity
        if action.entity.isa(Builtins.ValueType):
            ident = Var()
            actions.append(self.construct(action.entity.type, [ident, action.entity]))
        actions.append(self.call(Builtins.make_identity, [Var(value=params), ident]))
        actions.reverse()
        return actions

    def install(self, item:Type|Property):
        return self.call(Builtins.Install, [item])

    def construct(self, item:Type|Property, params:List[Any]):
        return self.relation_action(ActionType.Construct, item, params)

    def property_named(self, name:str, types:List[Type]):
        found = Property.find(name, types)
        if not found:
            ifound = Property.find(name, types, ignore_case=True)
            if ifound:
                from . import errors
                errors.PropertyCaseMismatch(name, ifound.name)
            found = Property(name, Builtins.Any)
        for t in types:
            if found not in t.properties:
                t.properties.append(found)
        return found

    def raw(self, code:str):
        code = textwrap.dedent(code)
        return self.relation_action(ActionType.Call, Builtins.RawCode, [code])

    def raw_task(self, code:str):
        return Task(behavior=Behavior.Sequence, items=[self.raw(textwrap.dedent(code))])

    def aggregate_def(self, op:str):
        return Task(name=op, parents=[Builtins.Aggregate], properties=[
            Property("group", Builtins.Any, True),
            Property("projection", Builtins.Any, True),
            Property("pre_args", Builtins.Any, True),
            Property("result", Builtins.Any),
        ])

    def eq(self, a:Any, b:Any, approx=False):
        if approx:
            return self.call(Builtins.approx_eq, [a, b])
        return self.call(Builtins.eq, [a, b])

#--------------------------------------------------
# Printer
#--------------------------------------------------

class Namer():
    def __init__(self, unnamed_vars=False):
        self.name_mapping = {}
        self.names = set()
        self.unnamed_vars = unnamed_vars

    def get_safe_name(self, name:str):
        name = name.lower()
        if name in self.names:
            ix = 2
            while f"{name}{ix}" in self.names:
                ix += 1
            name = f"{name}{ix}"
        self.names.add(name)
        return name

    def get(self, item:Var|Task|Type|Property):
        if item.id in self.name_mapping:
            return self.name_mapping[item.id]

        name = item.name if not self.unnamed_vars or not isinstance(item, Var) else None
        name = name or ("t" if isinstance(item, Task) else "v")
        name = self.get_safe_name(name)
        self.name_mapping[item.id] = name
        return name

    def reset(self):
        self.name_mapping.clear()
        self.names.clear()

class Printer():
    def __init__(self, unnamed_vars=False):
        self.indent = 0
        self.namer = Namer(unnamed_vars=unnamed_vars)

    def indent_str(self):
        return " " * 4 * self.indent

    def print(self, item:AllItems|Base|Value, is_sub=False):
        if isinstance(item, Task):
            return self.task(item, is_sub)
        elif isinstance(item, Type):
            return self.type(item, is_sub)
        elif isinstance(item, Property):
            return self.property(item, is_sub)
        elif isinstance(item, Agent):
            return self.agent(item, is_sub)
        elif isinstance(item, Var):
            return self.var(item, is_sub)
        elif isinstance(item, Action):
            return self.action(item, is_sub)
        elif isinstance(item, bytes):
            return base64.b64encode(item).decode()[:-2]
        elif isinstance(item, list):
            vs = [self.print(i, is_sub) for i in item]
            if len(item) > 20:
                return f"[{', '.join(vs[0:5])}, ... {', '.join(vs[-2:])}]"
            return f"[{', '.join(vs)}]"
        elif isinstance(item, str) or isinstance(item, bool) or isinstance(item, numbers.Number):
            return json.dumps(item)
        elif isinstance(item, datetime) or isinstance(item, date):
            return item.isoformat()
        return "UNKNOWN"
        # raise Exception(f"Unknown item type: {type(item)}")

    def type(self, type:Type, is_sub=False):
        return type.name

    def property(self, property:Property, is_sub=False):
        return property.name

    def task(self, task:Task, is_sub=False):
        self.indent += 1
        items = '\n'.join([self.print(i, is_sub) for i in task.items])
        self.indent -= 1
        behavior = task.behavior.value
        if is_sub:
            final = f"""{behavior}\n{items}\n"""
        else:
            final = f"""{self.indent_str()}{behavior}\n{items}\n"""
        return final

    def agent(self, agent:Agent, is_sub=False):
        return agent.name

    def var(self, var:Var, is_sub=False):
        if var.value is not None:
            if isinstance(var.value, Task):
                return "SUBTASK"
            elif isinstance(var.value, Type) or isinstance(var.value, Property):
                return str(var.value)
            return self.print(var.value, is_sub)
        return self.namer.get(var)

    def action(self, action:Action, is_sub=False):
        op = action.action.value
        entity_value = action.entity.value
        as_relation = False
        as_quantifier = False
        subs = []
        body = ""

        if entity_value == Builtins.Return:
            op = "return"
            as_relation = True
        elif isinstance(entity_value, Task):
            if op == "call" and len(entity_value.items):
                subs.append(entity_value)
            elif op == "call" and entity_value.isa(Builtins.Quantifier):
                as_quantifier = True
                subs.append(action.bindings[entity_value.properties[1]].value)
            else:
                as_relation = True
        elif isinstance(entity_value, Property) or isinstance(entity_value, Type):
            as_relation = True

        if as_relation:
            args = " ".join([self.print(v, is_sub) for _,v in action.bindings.items()])
            if op == "return":
                body = args
            else:
                entity_value = cast(Type, entity_value)
                rel_name = entity_value.name or self.namer.get(entity_value)
                body = f"{rel_name}({args})"
        elif as_quantifier:
            entity_value = cast(Task, entity_value)
            group_vars:Any = action.bindings[entity_value.properties[0]].value
            group = ", ".join([self.print(v) for v in group_vars])
            if group:
                body = f"{entity_value.name.lower()}({group}) "
            else:
                body = f"{entity_value.name.lower()} "
        elif not len(subs):
            types = [t.name for t in action.types]
            args = [f"{k.name}({self.print(v, is_sub)})" for k,v in action.bindings.items()]
            body = f"{self.print(action.entity)} | " + " ".join(types + args)

        final = f"{self.indent_str()}{op :>9} | {body}"
        if len(subs):
            self.indent += 1
            final += "\n".join([self.print(s, True) for s in subs])
            self.indent -= 1
        return final


#--------------------------------------------------
# Utils
#--------------------------------------------------

def _graph_debug(actions, in_degree, vars_provided_by):
    rich.print("\n[yellow bold]STRATIFY CHECK")
    for item in actions:
        print("-------------------------------------------")
        print(in_degree.get(item, 0), item)
        (requires, provides, refs) = item.requires_provides()
        if item.is_subtask_call():
            for ref in refs:
                if vars_provided_by.get(ref):
                    requires.add(ref)
        rich.print("    in:", ", ".join([str(r.id) for r in requires]))
        rich.print("    out:", ", ".join([str(p.id) for p in provides]))
        rich.print("    refs:", ", ".join([str(r.id) for r in refs]))

class Utils:

    @staticmethod
    def action_graph(actions:List[Action]):
        graph:Dict[Action, List[Action]] = defaultdict(list)
        in_degree:Dict[Action, int] = defaultdict(int)
        vars_provided_by:Dict[Var, List[Action]] = defaultdict(list)
        vars_required_by:Dict[Var, List[Action]] = defaultdict(list)

        item_refs = {}
        cut_vars = set()

        for item in actions:
            (requires, provides, refs) = item.requires_provides()
            if item.is_subtask_call():
                item_refs[item] = refs
            for var in provides:
                if var in cut_vars:
                    vars_required_by[var].append(item)
                else:
                    vars_provided_by[var].append(item)
            for var in requires:
                vars_required_by[var].append(item)

            if item.entity.isa(Builtins.Aggregate) or item.entity.isa(Builtins.Relation):
                cut_vars.update(provides)


        # ensure that subtasks depend on any var they reference in the outer scope
        for item, refs in item_refs.items():
            for ref in refs:
                if vars_provided_by.get(ref):
                    vars_required_by[ref].append(item)

        for var, requirers in vars_required_by.items():
            providers = vars_provided_by.get(var, [])
            for r in requirers:
                for provider in providers:
                    graph[provider].append(r)
                    in_degree[r] += 1

        # _graph_debug(actions, in_degree, vars_provided_by)

        return graph, in_degree

    #--------------------------------------------------
    # Stratify
    #--------------------------------------------------

    @staticmethod
    def stratify(actions:List[Action]):
        graph, in_degree = Utils.action_graph(actions)
        strata = []
        current_stratum = []
        handled = set()

        while True:
            no_dependencies = [item for item in actions if item not in handled and in_degree[item] == 0]

            if not no_dependencies:
                break

            current_stratum.extend(no_dependencies)
            for item in no_dependencies:
                handled.add(item)
                for dependent in graph[item]:
                    in_degree[dependent] -= 1

                del graph[item]
                del in_degree[item]

            strata.append(current_stratum)
            current_stratum = []

        if graph:
            # for item in graph:
            #     rich.print(f"[red bold]Unresolved dependency[/red bold]: {str(item)}")
            raise ValueError("There is a cycle in the graph!")

        return strata

    #--------------------------------------------------
    # Toposort
    #--------------------------------------------------

    @staticmethod
    def sort(actions:List[Action]):
        strata = Utils.stratify(actions)
        sorted = []
        for stratum in strata:
            stratum.sort(key=lambda x: x.id)
            for item in stratum:
                sorted.append(item)
        return sorted

    #--------------------------------------------------
    # Vars
    #--------------------------------------------------

    @staticmethod
    def gather_vars(task_items:List[Action], with_values=False) -> List[Var]:
        vars_seen = set()
        vars = []
        for item in task_items:
            item_vars = [item.entity, *item.bindings.values()]
            for v in item_vars:
                if isinstance(v.value, list):
                    for var in v.value:
                        if isinstance(var, Var) and (with_values or var.value is None):
                            if var not in vars_seen:
                                vars_seen.add(var)
                                vars.append(var)
                else:
                    if with_values or v.value is None:
                        if v not in vars_seen:
                            vars_seen.add(v)
                            vars.append(v)
        return vars

    @staticmethod
    def gather_task_vars(task:Task, with_values=False, only_direct_children=False, seen=None) -> Set[Var]:
        if seen is None:
            seen = set()
        if task in seen:
            return set()
        seen.add(task)

        refs = set()
        for i in task.items:
            vs = i.vars()
            for v in vs:
                if i.action == ActionType.Call and not only_direct_children and v.value and isinstance(v.value, Task):
                    refs.update(Utils.gather_task_vars(v.value, with_values, only_direct_children, seen))
                elif with_values or (v.value is None and not v.isa(Builtins.InlineRawData)):
                    refs.add(v)
        return refs



from .splinter import Splinter
from .list_types import RewriteListTypes
from .gc_nodes import GarbageCollectNodes
from .flatten import Flatten

__all__ = ["Splinter", "RewriteListTypes", "GarbageCollectNodes", "Flatten"]



from relationalai.early_access.lqp import ir as lqp
from relationalai.early_access.metamodel import ir

from hashlib import sha256

class UniqueNames:
    def __init__(self):
        # Track count of seen names
        self.seen = dict[str, int]()
        # Maps id to unique name
        self.id_to_name = dict[int,str]()

    def get_name(self, name: str) -> str:
        # TODO: Do we need to be threadsafe?
        if name in self.seen:
            self.seen[name] += 1
            id = self.seen[name]
            new_name = f"{name}_{id}"
            self.seen[new_name] = 1
            return new_name
        else:
            self.seen[name] = 1
            return f"{name}"

    # Get a unique name for the given id. If the id is already in the map, return the
    # existing name. Otherwise, generate a new name using the suggested_name and
    # store it in the map.
    def get_name_by_id(self, id: int, suggested_name:str) -> str:
        if id in self.id_to_name:
            return self.id_to_name[id]

        name = self.get_name(suggested_name)
        self.id_to_name[id] = name
        return name

class TranslationCtx:
    def __init__(self, model):
        # TODO: comment these fields
        self.unique_names = UniqueNames()
        self.id_to_orig_name = {}
        self.output_ids = []

def gen_unique_var(ctx: TranslationCtx, name_hint: str):
    """
    Generate a new variable with a unique name based on the provided hint.
    """
    name = ctx.unique_names.get_name(name_hint)
    return lqp.Var(name=name, meta=None)

def is_constant(arg, expected_type):
    """
    Check if the argument is a constant of the expected type.
    """
    if isinstance(arg, ir.Literal):
        return is_constant(arg.value, expected_type)

    return isinstance(arg, expected_type)

def rename_vars_var(var: lqp.Var, var_map: dict[str, lqp.Var]) -> lqp.Var:
    return var_map.get(var.name, var)

def rename_vars_relterm(term: lqp.RelTerm, var_map: dict[str, lqp.Var]) -> lqp.RelTerm:
    if isinstance(term, lqp.Var):
        return rename_vars_var(term, var_map)
    else:
        return term  # Constants do not change

def rename_vars_term(term: lqp.Term, var_map: dict[str, lqp.Var]) -> lqp.Term:
    if isinstance(term, lqp.Var):
        return rename_vars_var(term, var_map)
    else:
        return term  # Constants do not change

def rename_vars_abstraction(abstraction: lqp.Abstraction, var_map: dict[str, lqp.Var]) -> lqp.Abstraction:
    new_vars = [(var_map.get(var.name, var), typ) for (var, typ) in abstraction.vars]
    new_value = rename_vars_formula(abstraction.value, var_map)
    return lqp.Abstraction(vars=new_vars, value=new_value, meta=abstraction.meta)

def rename_vars_formula(formula: lqp.Formula, var_map: dict[str, lqp.Var]) -> lqp.Formula:
    if isinstance(formula, lqp.Primitive):
        return lqp.Primitive(
            name=formula.name,
            terms=[rename_vars_relterm(term, var_map) for term in formula.terms],
            meta=formula.meta
        )
    elif isinstance(formula, lqp.Atom):
        return lqp.Atom(
            name=formula.name,
            terms=[rename_vars_term(term, var_map) for term in formula.terms],
            meta=formula.meta
        )
    elif isinstance(formula, lqp.Not):
        return lqp.Not(arg=rename_vars_formula(formula.arg, var_map), meta=formula.meta)
    elif isinstance(formula, lqp.Exists):
        return lqp.Exists(body=rename_vars_abstraction(formula.body, var_map), meta=None)
    elif isinstance(formula, lqp.Reduce):
        return lqp.Reduce(
            op=formula.op,
            body=rename_vars_abstraction(formula.body, var_map),
            terms=[rename_vars_term(term, var_map) for term in formula.terms],
            meta=formula.meta
        )
    elif isinstance(formula, lqp.FFI):
        return lqp.FFI(
            meta=formula.meta,
            name=formula.name,
            args=[rename_vars_abstraction(arg, var_map) for arg in formula.args],
            terms=[rename_vars_term(term, var_map) for term in formula.terms]
        )
    elif isinstance(formula, lqp.Conjunction):
        return lqp.Conjunction(
            args=[rename_vars_formula(arg, var_map) for arg in formula.args],
            meta=formula.meta
        )
    elif isinstance(formula, lqp.Disjunction):
        return lqp.Disjunction(
            args=[rename_vars_formula(arg, var_map) for arg in formula.args],
            meta=formula.meta
        )
    else:
        raise NotImplementedError(f"Unknown formula type: {type(formula)}")

def lqp_hash(node: str) -> int:
    h = int.from_bytes(sha256(node.encode()).digest(), byteorder='big', signed=False)
    # Ensure it's within the 128-bit range
    return h % (2**128)



#
# The names used to refer to components of a schema. Always involves a name,
# zero or more dashes (' symbols in Z), and a prefix string that could be
# empty.
#
from relationalai.early_access.dsl.constants import PRIME


class CompName:

    # Generates a CompName from a name *nm* in context *ctx* in which the
    # component is declared and a level *i* that is between 0 and the depth
    # of ctx. The CompName at level i must reflect the cumulative decorations
    # (dashes and prefixes) that were applied at levels i .. d.
    #
    @staticmethod
    def gen_from_context(nm, ctx, i, dashes=0):

        dsh = ctx.dashes[i] + dashes
        pref = ctx.prefixes[i]
        return CompName(nm, dashes=dsh, prefix=pref)

    def __init__(self, nm, dashes=0, prefix=""):
        self.name = nm
        self.prefix = prefix
        self.dashes = dashes

    def display(self):
        if self.prefix == "":
            nm = self.name
        else:
            nm = self.name[0].capitalize() + self.name[1:]
        return self.prefix + nm + PRIME * self.dashes

    # This predicate returns True if this name is decorated by a single dash
    #
    def decorated(self) -> bool:
        return self.dashes == 1

    def undecorated(self) -> bool:
        return self.dashes == 0

    def undash(self):
        dashes = self.dashes - 1
        if dashes < 0:
            raise Exception(f"++ Cannot undash {self.display()}")
        else:
            return CompName(self.name, dashes, self.prefix)

    def entityid(self):
        return hash(self.display())



import dataclasses
import datetime
from abc import abstractmethod, ABC
from enum import Enum
from typing import Optional, Any, TypeVar, Generic, Tuple
from typing import Union

import relationalai.early_access.builder as qb
from relationalai.early_access.builder import annotations
from relationalai.early_access.dsl.orm.relationships import Relationship, Role
from relationalai.early_access.dsl.orm.types import Concept
from relationalai.early_access.dsl.snow.common import ColumnRef, ForeignKey

PrimitiveFilterBy = Union[int, float, str, bool, datetime.date, datetime.datetime, Enum]
FilterBy = Union[PrimitiveFilterBy, qb.Expression, tuple[qb.Expression]]

# defining symbol in codegen to not expose this to the user
Symbol = Concept.builtins["Symbol"]


class BindableTable(ABC):
    """
    A class representing a bindable table.
    """

    @abstractmethod
    def key_type(self) -> qb.Concept:
        pass

    @abstractmethod
    def physical_name(self) -> str:
        pass


class BindableAttribute(ABC):

    @property
    @abstractmethod
    def table(self) -> BindableTable:
        pass

    @abstractmethod
    def physical_name(self) -> str:
        pass

    @abstractmethod
    def type(self) -> qb.Concept:
        pass

    @property
    @abstractmethod
    def references_column(self) -> Optional[ColumnRef]:
        pass

    @abstractmethod
    def column_ref(self) -> ColumnRef:
        pass

    @abstractmethod
    def decimal_scale(self) -> Optional[int]:
        pass

    @abstractmethod
    def decimal_size(self) -> Optional[int]:
        pass


@dataclasses.dataclass(frozen=True)
class Binding:
    """
    Base class for all bindings in the model. It captures the column and the optional filter that should be applied
    to the binding.

    Examples of filters:
        - `FilterBy` can be a primitive value, such as an integer or a string, which will be used to filter the
          values in the column.
        - An expression can be used to filter the values in the column based on a more complex condition.
        - A tuple of expressions can be used to filter the values in the column based on multiple expressions, joined
          together.

        Note: all the expressions must be atomic, i.e., they cannot contain arbitrarily nested formulas.

        # binds the column to the role in the relationship, filtering out any rows that don't match the specified value
        {source}.COLUMN.binds(MyRelationship, filter_by="some_value")

        # binds the column to the role in the relationship, filtering out any rows that don't match the specified
        # expression
        {source}.COLUMN.binds(MyRelationship, filter_by=(
            {source}.ANOTHER_COLUMN >= 18
        ))

        # multiple atoms can be used too, resulting in a conjunction
        {source}.COLUMN.binds(MyRelationship[MyRole], filter_by=(
            {source}.ANOTHER_COLUMN >= 18,
            {source}.YET_ANOTHER_COLUMN == "some_value"
        ))
    """
    column: 'BindableColumn'
    filter_by: Optional[FilterBy]


@dataclasses.dataclass(frozen=True)
class RoleBinding(Binding):
    """
    RoleBinding represents a binding between a column and a role in a relationship.

    See `BindableColumn.binds` for more details on how to use this binding.
    """
    role: Role

    def __str__(self):
        return f'RoleBinding[{self.column.table.physical_name()}:{self.column.physical_name()} -> {self.role.player().name()}]'


@dataclasses.dataclass(frozen=True)
class AbstractConceptBinding(Binding):
    """
    Represents a parent class for all concept bindings. Doesn't have any specific semantics but serves as a base
    class for this class hierarchy.
    """
    entity_type: qb.Concept


@dataclasses.dataclass(frozen=True)
class IdentifierConceptBinding(AbstractConceptBinding):
    """
    Represents a binding between an identifier column and a specific entity type.

    This binding could either represent a constructor binding (instances of the entity type are constructed from the
    values), referent binding (instances of the entity type are being looked up by the values), or a subtype binding
    (instances are being looked up using the parent type's ref scheme, or it acts as a constructor for the subtype).

    See `BindableColumn.identifies` and `BindableColumn.references` for more details on how to use this binding.
    """

    def __str__(self):
        return f'IdentifierConceptBinding[{self.column.table.physical_name()}:{self.column.physical_name()} -> {self.entity_type.name()}]'


@dataclasses.dataclass(frozen=True)
class ReferentConceptBinding(AbstractConceptBinding):
    """
    Represents a binding between an identifier column and a specific entity type, where the values in the column
    are used to look up existing instances of the entity type.

    See `BindableColumn.identifies` and `BindableColumn.references` for more details on how to use this binding.
    """

    def __str__(self):
        return f'ReferentConceptBinding[{self.column.table.physical_name()}:{self.column.physical_name()} -> {self.entity_type.name()}]'


@dataclasses.dataclass(frozen=True)
class PartitioningConceptBinding(AbstractConceptBinding):
    """
    Represents a binding between a column and a subtype of an entity type, where the values in the column
    are used to filter existing instances of the parent entity type; by filtering out the rows that do not match
    the specified value.

    See `BindableColumn.filters_subtype` for more details on how to use this binding.
    """
    has_value: PrimitiveFilterBy

    def __str__(self):
        return f'PartitioningConceptBinding[{self.column.table.physical_name()}:{self.column.physical_name()} == {self.has_value}]'


class BindableColumn(BindableAttribute, qb.Relationship, ABC):
    _source: BindableTable
    _references_column: Optional[ColumnRef]

    def __init__(self, col_name: str, col_type: qb.Concept, source: BindableTable, model):
        madlib, short_name = self._handle_params(col_type, source)
        qb.Relationship.__init__(self, madlib, model=model.qb_model(), short_name=short_name)
        self.annotate(annotations.external)
        self._col_name = col_name
        self._type = col_type
        self._source = source
        self._orm_model = model
        self._references_column = None

    @staticmethod
    def _handle_params(col_type: qb.Concept, source: BindableTable):
        """
        Handles the construction of Relationship parameters.

        Relationship name is the physical name of the source, as it's also specialized.
        We use separate QB relationships, as this allows proper typing, but in core Rel
        it's treated as a single relation; hence the name is the same.
        """
        short_name = source.physical_name()
        madlib = f"{short_name} {{Symbol}} {{Integer}} {{{col_type}}}"
        return madlib, short_name

    def __call__(self, *args):
        """
        Allows the column to be called as a function, which is handy for manual data weaving.

        Example:
            row = Integer.ref()
            where(
                {source}.ID(row, id),
                {source}.NAME(row, name),
                person := Person.new(id=id)
            ).define(
                person,
                Person.name(person, name)
            )
        """
        if len(args) == 2:
            args = [Symbol(self.physical_name()), *args]
        if len(args) != self._arity():
            raise ValueError(f'Expected {self._arity()} arguments, got {len(args)}')
        return qb.Relationship.__call__(self, *args)

    def identifies(self, entity_type: Concept, filter_by: Optional[FilterBy] = None):
        """
        Binds a column that identifies an entity type. With that, the values in the column are used to construct
        instances of the entity type.

        Examples:
            {source}.ID.identifies(Person) # binds the column values to the Person entity type, constructing instances
        """
        if isinstance(entity_type, Enum):
            raise ValueError(f'Cannot bind to Enum {entity_type}, use `references` instead')
        binding = IdentifierConceptBinding(column=self, entity_type=entity_type, filter_by=filter_by)
        self._orm_model.binding(binding)

    def references(self, entity_type: Concept, filter_by: Optional[FilterBy] = None):
        """
        Binds a column that references an identifier of an entity type. With that, the values in the column are
        used to look up existing instances of the entity type.

        Examples:
            {source}.ID.identifies(Person) # binds the column values to the Person entity type, constructing instances
            {another_source}.PERSON_ID.references(Person) # binds the column values to the Person entity type,
                                                          # looking up existing instances

            {source}.TRACKING_NUMER.identifies(TrackingNumber) # constructs instances of the TrackingNumber entity type
            {source}.RETURN_TRACKING_NUMBER.references(TrackingNumber) # looks up existing instances of the TrackingNumber entity type
        """
        binding = ReferentConceptBinding(column=self, entity_type=entity_type, filter_by=filter_by)
        self._orm_model.binding(binding)

    def filters_subtype(self, sub_type: Concept, by_value: Any):
        """
        Binds a column that partitions some parent entity type into a subtype. The values in the column are used to filter
        existing instances of the parent entity type by filtering out the rows that do not match the
        specified value.

        Examples:
            {source}.CORPORATE_ACTION_ID.identifies(CorporateAction) # constructs instances of the CorporateAction entity type
            {source}.CORPORATE_ACTION_TYPE.filters_subtype(Split, by_value="SPLIT") # partitions CorporateAction into Split subtype
            {source}.CORPORATE_ACTION_TYPE.filters_subtype(Merger, by_value="MERGER") # partitions CorporateAction into Merger subtype
        """
        binding = PartitioningConceptBinding(column=self, entity_type=sub_type, has_value=by_value, filter_by=None)
        self._orm_model.binding(binding)

    def binds(self, elm, filter_by: Optional[FilterBy] = None):
        """
        Binds the column to a role in a relationship or a role itself. It's possible to bind a column to a binary
        relationship, which will bind to the last role in the relationship.

        Examples:
            {source}.{column}.binds(MyRelationship) # binds the last role in MyRelationship

            {source}.{column}.binds(MyRelationship[MyConcept]) # binds the role played by MyConcept (if unique)

            # binding by index (may be useful for relationships with multiple roles played by the same concept)
            # BecameFriendsAt = Relationship("{Person} at {DateTime} befriended {Person}")
            friends_source.PERSON_ID.binds(BecameFriendsAt[0]) # binds the first Person role
            friends_source.BEFRIENDED_AT.binds(BecameFriendsAt[DateTime]) # binds the DateTime role
            friends_source.FRIEND_ID.binds(BecameFriendsAt[2]) # binds the second Person role (third in the relationship)
        """
        if isinstance(elm, Relationship):
            # this binds to the last role in binary relations
            if elm._arity() > 2:
                raise ValueError(f'Expected binary or unary relationship, got arity {elm._arity()}')
            roles = elm._roles()
            role = roles[-1]
        elif isinstance(elm, Role):
            role = elm
        else:
            raise Exception(
                f'Expected ORM Relationship or Role, got {type(elm)} - QB Relationships cannot be used in bindings')
        binding = RoleBinding(role=role, column=self, filter_by=filter_by)
        self._orm_model.binding(binding)

    @property
    def table(self):
        return self._source

    @property
    def references_column(self) -> Optional[ColumnRef]:
        return self._references_column

    @references_column.setter
    def references_column(self, ref: ColumnRef):
        self._references_column = ref

    def physical_name(self) -> str:
        return self._col_name

    def type(self) -> qb.Concept:
        return self._type

    def column_ref(self) -> ColumnRef:
        return ColumnRef(self._source.physical_name(), self.physical_name())


T = TypeVar("T", bound=BindableColumn)
class AbstractBindableTable(BindableTable, Generic[T], ABC):
    _foreign_keys: set[ForeignKey]

    def __init__(self, name: str, model, foreign_keys: set[ForeignKey]):
        super().__init__()
        self._table = name
        self._model = model
        self._cols: dict[str, T] = dict()
        self._foreign_keys = foreign_keys
        self._generate_declare()

    def _generate_declare(self):
        src = f"declare {self.physical_name()}"
        self._model.qb_model().define(qb.RawSource('rel', src))

    def __getattr__(self, key):
        if key in self.__dict__:
            return self.__dict__[key]
        if key in self._cols:
            return self._cols[key]
        raise AttributeError(f'Table "{self.physical_name()}" has no column named "{key}"')

    def foreign_key(self, *refs: Tuple[BindableColumn, BindableColumn]):
        source_columns = []
        target_columns = []

        for source, target in refs:
            source_columns.append(ColumnRef(source.table.physical_name(), source.physical_name()))
            target_columns.append(ColumnRef(target.table.physical_name(), target.physical_name()))

        source_col_names = "_".join(col.column for col in source_columns)
        target_col_names = "_".join(col.column for col in target_columns)

        fk_name = f"fk_{source_col_names}__to__{target_col_names}"

        fk = ForeignKey(fk_name, source_columns, target_columns)
        self._foreign_keys.add(fk)
        self._process_foreign_key(fk)

    def key_type(self) -> qb.Concept:
        return qb.Integer

    def columns(self):
        return self._cols

    def _process_foreign_keys(self):
        for fk in self._foreign_keys:
            self._process_foreign_key(fk)

    def _process_foreign_key(self, fk):
        # TODO : this doesn't work for composite FKs
        for col in fk.source_columns:
            target_col = fk.target_columns[0]
            self._cols[col.column].references_column = target_col



import typing

from relationalai.util.graph import topological_sort
from relationalai.early_access.dsl.orm.constraints import Unique, Mandatory, RoleValueConstraint, \
    InclusiveSubtypeConstraint, ExclusiveSubtypeConstraint, Range, RingConstraint
from relationalai.early_access.dsl.orm.models import Model
from relationalai.early_access.dsl.orm.relationships import Role
from relationalai.early_access.dsl.orm.types import Concept
from relationalai.early_access.metamodel.util import Printer as BasePrinter, OrderedSet
import relationalai.early_access.builder as qb

class Printer(BasePrinter):
    _unique_constraints: OrderedSet[Unique]
    _mandatory_constraints: OrderedSet[Mandatory]
    _role_value_constraints: OrderedSet[RoleValueConstraint]
    _inclusive_subtype_constraints: OrderedSet[InclusiveSubtypeConstraint]
    _exclusive_subtype_constraints: OrderedSet[ExclusiveSubtypeConstraint]
    _ring_constraints: OrderedSet[RingConstraint]

    def __init__(self, io: typing.Optional[typing.IO[str]] = None):
        # Set the base class field (frozen)
        object.__setattr__(self, 'io', io)

        # Set all mutable fields using object.__setattr__
        object.__setattr__(self, '_unique_constraints', OrderedSet())
        object.__setattr__(self, '_mandatory_constraints', OrderedSet())
        object.__setattr__(self, '_role_value_constraints', OrderedSet())
        object.__setattr__(self, '_inclusive_subtype_constraints', OrderedSet())
        object.__setattr__(self, '_exclusive_subtype_constraints', OrderedSet())
        object.__setattr__(self, '_ring_constraints', OrderedSet())


    def to_string(self, model: Model, enums_enabled: bool=False, pyi_enabled: bool=False) -> None:
        self._process_constraints(model)

        self._print_nl("import relationalai.early_access.builder as qb")
        self._print_nl("from relationalai.early_access.dsl.orm.models import Model")
        if pyi_enabled:
            self._print_nl("from relationalai.early_access.dsl.orm.models import Concept")
        self._print_nl("from relationalai.early_access.dsl.orm.constraints import Range")
        self._print_nl("from relationalai.early_access.dsl.orm.constraints import RingType")
        self._nl()

        self._print_nl(f"model = Model(name='{model.name}', is_primary={model.is_primary}"
                       f"{self._print_if_true('dry_run', model.qb_model()._dry_run)})")
        # use class stubs for new Concepts which extend orm QB Concept
        if not pyi_enabled:
            self._print_nl("Concept = model.Concept")
        if enums_enabled:
            self._print_nl("Enum = model.Enum")
        self._print_nl("Relationship = model.Relationship")
        self._print_nl("Unique = model.unique")
        self._print_nl("Mandatory = model.mandatory")
        self._print_nl("RoleValueConstraint = model.role_value_constraint")
        self._print_nl("InclusiveSubtypeConstraint = model.inclusive_subtype_constraint")
        self._print_nl("ExclusiveSubtypeConstraint = model.exclusive_subtype_constraint")
        self._print_nl("RingConstraint = model.ring")

        self._nl()
        self._handle_concepts(model, enums_enabled, pyi_enabled)

        self._nl()
        self._handle_relationships(model)

        self._nl()
        self._handle_ref_schemes(model)

        self._nl()
        self._handle_uniqueness_constraints()

        self._nl()
        self._handle_mandatory_constraints()

        self._nl()
        self._handle_role_value_constraints()

        self._nl()
        self._handle_subtype_constraints()

        self._nl()
        self._handle_ring_constraints()

    def _handle_concepts(self, model: Model, enums_enabled: bool, pyi_enabled: bool) -> None:
        concepts_map = model.concepts_map()
        enums_map = model.enums_map()
        sorted_concepts = self._sort_dependency_graph(model.concepts())
        # print Concept classes
        if pyi_enabled:
            for name in sorted_concepts:
                c = concepts_map.get(name)
                if c is None:
                    raise ValueError(f"The concept '{name}' was not declared but used as concept domain.")
                # skip enums when they are enabled for printer
                if c._is_enum() and enums_enabled:
                    continue
                self._print_nl(f"class {name}Concept(Concept):")
                self._print_nl(f'\t"""Represents {name} concept."""')
            self._nl()
        for name in sorted_concepts:
            c = concepts_map.get(name)
            if c is None:
                raise ValueError(f"The concept '{name}' was not declared but used as concept domain.")
            if enums_enabled and c._is_enum():
                self._print_nl(f"{name} = Enum('{name}', {[e.name for e in enums_map[name]]})")
            else:
                extends = c._extends
                if not enums_enabled and c._is_enum():
                    # todo: derive enum type
                    # ORM adapter produces only string enums
                    extends_elements = ["qb.String"]
                else:
                    extends_elements = [f"{self._get_type(ext)}" for ext in extends]
                extends_str = f", extends=[{', '.join(extends_elements)}]" if extends else ""
                concept_class = name + 'Concept' if pyi_enabled else 'Concept'
                params = f"({'model, ' if pyi_enabled else ''}'{name}'{extends_str})"
                self._print_nl(f"{name} = {concept_class}{params}")

    def _handle_relationships(self, model: Model) -> None:
        for rel in model.relationships():
            # skip autogenerated 'name' Relationship for Enums
            if rel._name == 'name' and rel._parent is not None and rel._parent._is_enum():
                continue
            # print a root Relationship
            self._print_nl(f"{self._get_relationship_name(rel)} = "
                           f"Relationship('{rel._madlib}'{self._print_if_not_empty_and_not_equal('short_name', rel._passed_short_name, rel._name)})")
            # print remaining RelationshipReadings if any
            for r in rel._readings[1:]:
                self._print_nl(f"{self._get_relationship_name(r)} = "
                               f"{self._get_relationship_name(rel)}.alt('{r._madlib}'{self._print_if_not_empty_and_not_equal('short_name', r._passed_short_name, r._name)})")

    def _handle_ref_schemes(self, model: Model) -> None:
        for concept_name, concept in model.entity_types_map().items():
            if not isinstance(concept, Concept) or not concept._reference_schemes:
                continue

            for ref_scheme in concept._reference_schemes:
                rel_names = [self._get_relationship_name(rel) for rel in ref_scheme]
                self._print_nl(f"{concept_name}.identify_by({', '.join(rel_names)})")

    def _handle_uniqueness_constraints(self) -> None:
        for constraint in self._unique_constraints:
            if not constraint.is_preferred_identifier:
                elements = []
                for role in constraint.roles():
                    elements.append(self._get_role_name(role))
                self._print_nl(f"Unique({', '.join(elements)})")

    def _handle_mandatory_constraints(self):
        for constraint in self._mandatory_constraints:
            role = constraint.roles()[0]
            role_name = self._get_role_name(role)
            self._print_nl(f"Mandatory({role_name})")

    def _handle_role_value_constraints(self) -> None:
        for constraint in self._role_value_constraints:
            role = constraint.roles()[0]
            elements = []

            for value in constraint.values():
                if isinstance(value, Range):
                    if value._start is not None and value._end is not None:
                        elements.append(f"Range.between({value._start}, {value._end})")
                    elif value._end is not None:
                        elements.append(f"Range.to_value({value._end})")
                    else:
                        elements.append(f"Range.from_value({value._start})")
                elif isinstance(value, str):
                    elements.append(repr(value))
                else:
                    elements.append(str(value))

            role_name = self._get_role_name(role)
            self._print_nl(f"RoleValueConstraint({role_name}, [{', '.join(elements)}])")

    def _handle_subtype_constraints(self):
        self._emit_subtype_constraints(self._inclusive_subtype_constraints, "InclusiveSubtypeConstraint")
        self._nl()
        self._emit_subtype_constraints(self._exclusive_subtype_constraints, "ExclusiveSubtypeConstraint")

    def _handle_ring_constraints(self):
        for constraint in self._ring_constraints:
            elements = []
            for role in constraint.roles():
                elements.append(self._get_role_name(role))
            cst_types = []
            for tp in constraint.types:
                cst_types.append(f"RingType.{tp.name}")
            self._print_nl(f"RingConstraint([{', '.join(cst_types)}], {', '.join(elements)})")

    def _emit_subtype_constraints(self, constraints, constraint_type: str):
        for constraint in constraints:
            concepts = [self._get_type(c) for c in constraint.concepts()]
            self._print_nl(f"{constraint_type}({', '.join(concepts)})")

    def _process_constraints(self, model: Model) -> None:
        for c in model.constraints():
            if isinstance(c, Unique):
                self._unique_constraints.add(c)
            elif isinstance(c, Mandatory):
                self._mandatory_constraints.add(c)
            elif isinstance(c, RoleValueConstraint):
                self._role_value_constraints.add(c)
            elif isinstance(c, InclusiveSubtypeConstraint):
                self._inclusive_subtype_constraints.add(c)
            elif isinstance(c, ExclusiveSubtypeConstraint):
                self._exclusive_subtype_constraints.add(c)
            elif isinstance(c, RingConstraint):
                self._ring_constraints.add(c)


    def _get_role_name(self, role: Role) -> str:
        return f"{self._get_relationship_name(role._relationship)}['{role._field_ref._name}']"

    def _sort_dependency_graph(self, concepts: list[qb.Concept]):
        nodes = []
        edges = []
        for concept in concepts:
            name = concept._name
            nodes.append(name)
            for ext in concept._extends:
                if not self._is_builtin_type(ext):
                    edges.append((ext._name, name))
        return topological_sort(nodes, edges)

    @staticmethod
    def _get_relationship_name(rel: typing.Union[qb.Relationship, qb.RelationshipReading]) -> str:
        if rel._parent and rel._short_name:
            parent_name = rel._parent._name
            prefix = "qb." if parent_name in Concept.builtins else ""
            return f"{prefix}{parent_name}.{rel._short_name}"
        return rel._name

    def _get_type(self, t: Concept) -> str:
        name = self._get_concept_name(t)
        return f"qb.{name}" if name in Concept.builtins else name

    def _is_builtin_type(self, t: Concept) -> bool:
        return self._get_concept_name(t) in Concept.builtins

    @staticmethod
    def _get_concept_name(t: Concept) -> str:
        return 'Integer' if t._name == 'Int' else t._name

    @staticmethod
    def _print_if_not_empty(label: str, value: typing.Optional[str]) -> str:
        return f", {label}='{value}'" if value else ""

    @staticmethod
    def _print_if_not_empty_and_not_equal(label: str, value: typing.Optional[str], compare_to: typing.Optional[str]) -> str:
        return f", {label}='{value}'" if value and value != compare_to else ""

    @staticmethod
    def _print_if_true(label: str, value: bool) -> str:
        return f", {label}=True" if value else ""

EXCLUDED_RELATIONSHIPS = ["shape"]

class InterfacePrinter(BasePrinter):

    def __init__(self, io: typing.Optional[typing.IO[str]] = None):
        # Set the base class field (frozen)
        object.__setattr__(self, 'io', io)

    def to_string(self, model: Model, enums_enabled: bool=False) -> None:
        if enums_enabled:
            self._print_nl("from enum import Enum")
        self._print_nl("from typing import Union")
        self._print_nl("import relationalai.early_access.builder as qb")
        self._print_nl("from relationalai.early_access.dsl.orm.models import Concept")
        self._nl()
        if enums_enabled:
            self._handle_enums(model)
        self._handle_concepts(model, enums_enabled)

    def _handle_enums(self, model: Model) -> None:
        for e in model.enums():
            self._print_nl(f"class {e.__name__}(Enum):")
            for member in e:
                self._print_nl(f"\t{member.name}: {e.__name__}")
            self._print_nl("\n".join(self._get_declared_relationships(e._concept, EXCLUDED_RELATIONSHIPS))) #type: ignore
            self._nl()

    def _handle_concepts(self, model: Model, enums_enabled) -> None:
        for c in model.concepts():
            if enums_enabled and c._is_enum():
                continue
            exclude_list = list(EXCLUDED_RELATIONSHIPS)
            # when enum is declared but enums are disabled print it as a Concept without "name" Relationship
            if c._is_enum():
                exclude_list.append("name")
            self._print_nl(f"class {c._name}Concept(Concept):")
            rel_strs = self._get_declared_relationships(c, exclude_list)
            self._print_nl("\tpass" if len(rel_strs) == 0 else "\n".join(rel_strs))
            self._print_nl(f"{c._name}: {c._name}Concept")
            self._nl()

    def _get_declared_relationships(self, c: qb.Concept, exclude_list: list[str]) -> list[str]:
        rel_strs = []
        for rel in c._relationships.keys():
            if rel not in exclude_list:
                rel_strs.append(f"\t{rel}: Union[qb.Relationship, qb.RelationshipReading]")
        return rel_strs


from typing import Optional

from relationalai.early_access.dsl import ExternalRelation, EntityPopulationRelation, Namespace
from relationalai.early_access.dsl.core.exprs import _
from relationalai.early_access.dsl.core.relations import RelationSignature, rule, EntityInstanceRelation
from relationalai.early_access.dsl.core.rules import Rule, Vars
from relationalai.early_access.dsl.core.types.standard import standard_value_types, Hash
from relationalai.early_access.dsl.types import AbstractConcept
from relationalai.early_access.dsl.core.types import Type as CoreType


class AbstractEntityType(AbstractConcept, CoreType):

    def __init__(self, model, name):
        super().__init__(model, name)
        CoreType.__init__(self, name)
        self._poprel = None

    def population(self):
        if self._poprel is None:
            self._poprel = EntityPopulationRelation(self.namespace(), self.display(), RelationSignature(self))
        return self._poprel

    def ctor_name(self): return f"^{self.display()}"

    def entity(self): return True

    def addrule(self, rule):
        relation = self.population()
        self._model._add_relation(relation)
        if not isinstance(rule, Rule):
            # Assume that rule is a Python function that, when invoked can build a rule
            rule = Rule(relation).elaborate(rule)
        relation._rules.append(rule)
        return rule

    def entity_instance_relation(self, name:str, namespace: Optional[Namespace]=None):
        rel = EntityInstanceRelation(namespace, name, RelationSignature(self))
        self._add_relation(rel)
        return rel

    @staticmethod
    def root_unconstrained_type():
        return standard_value_types['Hash']

class EntityType(AbstractEntityType):

    def __init__(self, model, nm, *args, ref_schema_name: Optional[str] = None):
        super().__init__(model, nm)
        self._domain = [*args]
        self._constructor = None
        self._ref_schema_name = ref_schema_name

    def __xor__(self, args):
        ctr = self.constructor()
        if ctr is None:
            raise Exception(
                f"Entity '{self.display()}' domain is empty or this Entity is not a subtype of another EntityType with constructor")
        return ctr(*args)

    def build_relation_variable(self, args, kwargs):
        if len(args) == 0:
            raise Exception(f"Unexpected error in EntityType {self.display()}")

        if len(args) > 1:
            raise Exception(f"Cannot declare type variable of EntityType {self.display()} by supplying more than a type")

        if len(kwargs) != 0:
            raise Exception(
                f"Cannot use keyword arguments when instantiating a type parameter for EntityType {self.display()}")

        type = args[0]
        self._domain.append(type)
        return type

    def build_scalar_variable(self, args, kwargs):
        if len(args) == 0:
            raise Exception(f"Unexpected error in EntityType {self.display()}")

        if len(args) > 1:
            raise Exception(f"Cannot declare type variable of EntityType {self.display()} by supplying more than a type")

        if len(kwargs) != 0:
            raise Exception(
                f"Cannot use keyword arguments when instantiating a type parameter for EntityType {self.display()}")

        type = args[0]
        self._domain.append(type)
        return type

    # An EntityTypeConstructor is an n-ary functional Relation that maps tuples of
    # arity (n-1) to instances of the EntityType.
    def constructor(self):
        if self._constructor is None:
            # Create a constructor function (relation) when this EntityType is created
            ctortype = self.domain().copy()
            ctortype.append(self)
            self._constructor = ExternalRelation(self.namespace(),
                                                 self.ctor_name(),
                                                 RelationSignature(*tuple(ctortype)))
        return self._constructor

    def first(self, relation):
        with self:
            @rule()
            def r(e):
                relation(e, _)

    def filtered_by(self, filter, value):
        with self:
            @rule()
            def r(e):
                v = Vars(Hash)
                #
                value(v)
                filter(e, v)

    def qualified_name(self):
        return self._name

    def domain(self):
        return self._domain

    def is_composite(self):
        return len(self._domain) > 1

    def ref_schema_name(self):
        return self._ref_schema_name

    def pprint(self):
        domain_str = ", ".join([f"x{i} in {t.display()}" for i, t in enumerate(self._domain)])
        return f"entity type {self.display()}({domain_str})"

    def __str__(self):
        return f'EntityType({self.display()})'



from io import StringIO

from relationalai.early_access.dsl.adapters.orm.adapter import ORMAdapter
from relationalai.early_access.dsl.adapters.orm.adapter_qb import ORMAdapterQB
from relationalai.early_access.dsl.adapters.owl.adapter import OwlAdapter
from relationalai.early_access.dsl.ontologies.python_printer import PythonPrinter
from relationalai.early_access.dsl.orm.printer import Printer, InterfacePrinter

def orm_to_python(orm_file_path: str, output_file_path: str) -> None:
    model_to_python(ORMAdapter(orm_file_path).model, output_file_path)

def orm_to_python_qb(orm_file_path: str, output_file_path: str, enums_enabled: bool=False, pyi_enabled: bool=False) -> None:
    model = ORMAdapterQB(orm_file_path).model
    model_to_python_qb(model, output_file_path, enums_enabled, pyi_enabled)
    if pyi_enabled:
        model_to_python_interface_qb(model, f"{output_file_path}i", enums_enabled)

def owl_to_python(owl_file_path: str, output_file_path: str) -> None:
    model_to_python(OwlAdapter(owl_file_path).model, output_file_path)

def model_to_python(model, output_file_path: str) -> None:
    with StringIO() as s:
        PythonPrinter(s).to_python_string(model)
        with open(output_file_path, "w", encoding="utf-8") as f:
            f.write(s.getvalue())

def model_to_python_qb(model, output_file_path: str, enums_enabled: bool=False, pyi_enabled: bool=False) -> None:
    with StringIO() as s:
        Printer(s).to_string(model, enums_enabled, pyi_enabled)
        with open(output_file_path, "w", encoding="utf-8") as f:
            f.write(s.getvalue())

def model_to_python_interface_qb(model, output_file_path: str, enums_enabled: bool=False) -> None:
    with StringIO() as s:
        InterfacePrinter(s).to_string(model, enums_enabled)
        with open(output_file_path, "w", encoding="utf-8") as f:
            f.write(s.getvalue())

def orm_to_python_string(orm_file_path: str) -> str:
    return model_to_python_string(ORMAdapter(orm_file_path).model)

def orm_to_qb_python_string(orm_file_path: str, enums_enabled: bool=False, pyi_enabled: bool=False) -> str:
    return model_to_qb_python_string(ORMAdapterQB(orm_file_path).model, enums_enabled, pyi_enabled)

def owl_to_python_string(owl_file_path: str) -> str:
    return model_to_python_string(OwlAdapter(owl_file_path).model)

def model_to_python_string(model) -> str:
    with StringIO() as s:
        PythonPrinter(s).to_python_string(model)
        return s.getvalue()

def model_to_qb_python_string(model, enums_enabled: bool=False, pyi_enabled: bool=False) -> str:
    with StringIO() as s:
        Printer(s).to_string(model, enums_enabled, pyi_enabled)
        return s.getvalue()

def model_to_python_interface_string(model, enums_enabled: bool=False) -> str:
    with StringIO() as s:
        InterfacePrinter(s).to_string(model, enums_enabled)
        return s.getvalue()



from __future__ import annotations
import sys
from typing import Any

from ..clients.config import Config
from .base import SessionEnvironment
from .ipython import IPythonEnvironment

class HexEnvironment(IPythonEnvironment, SessionEnvironment):
    remote = True

    @classmethod
    def detect(cls):
        return "hex" in sys.modules or "hex_data_service" in sys.modules or "hex_api" in sys.modules

    def configure_session(self, config: Config, session: Any | None = None):
        # @NOTE: I have no idea why the exception when user is present exists,
        #        it was copied verbatim from the earlier session setup logic.
        if not session and config.get("user", None) is None:
            raise ValueError("A Session object should be provided when running in Hex. Import `hextoolkit` and supply `connection=hextoolkit.get_data_connection('<your connection name>').get_snowpark_session()` as an argument to `Model`.")

        return session



from relationalai.early_access.builder import Fragment
from relationalai.early_access.rel import Compiler as RelCompiler
from relationalai.early_access.metamodel import ir
from relationalai.early_access.devmode import Compiler as DevModeCompiler

from dataclasses import dataclass
from typing import Optional
import inspect
import tempfile
import webbrowser
import os
import html

@dataclass
class CompilationResult:
    python_source: str
    ir: str
    rel: Optional[str]
    sql: Optional[str]

    def show(self):
        """
        Displays the compilation results in a browser by creating a temporary HTML file.

        Args:
            compilation_result: Dictionary containing 'ir', 'rel', and 'sql' keys
                               as returned by the compile() function

        Returns:
            Path to the temporary HTML file
        """

        # Create HTML content with syntax highlighting using highlight.js
        html_content = """
        <!DOCTYPE html>
        <html>
        <head>
            <title>RAI Compilation Results</title>
            <link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.7.0/styles/default.min.css">
            <script src="https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.7.0/highlight.min.js"></script>
            <script src="https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.7.0/languages/python.min.js"></script>
            <script src="https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.7.0/languages/ruby.min.js"></script>
            <script src="https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.7.0/languages/sql.min.js"></script>
            <style>
                body {{ font-family: Arial, sans-serif; margin: 20px; }}
                .section {{ margin-bottom: 30px; }}
                h2 {{ color: #333; }}
                pre {{ background-color: #f5f5f5; padding: 10px; border-radius: 5px; overflow: auto; }}
                .error {{ color: red; }}
            </style>
        </head>
        <body>
            <h1>RAI Compilation Results</h1>

            <div class="section">
                <h2>Python Source</h2>
                <pre><code class="language-python">{}</code></pre>
            </div>

            <div class="section">
                <h2>IR (Intermediate Representation)</h2>
                <pre><code class="language-json">{}</code></pre>
            </div>

            <div class="section">
                <h2>Rel</h2>
                <pre><code class="language-ruby">{}</code></pre>
            </div>

            <div class="section">
                <h2>SQL</h2>
                <pre><code class="language-sql">{}</code></pre>
            </div>

            <script>hljs.highlightAll();</script>
        </body>
        </html>
        """.format(
            html.escape(self.python_source),
            html.escape(self.ir),
            html.escape(self.rel or "not available"),
            html.escape(self.sql or "not available"),
        )

        # Create a temporary HTML file
        with tempfile.NamedTemporaryFile(suffix='.html', delete=False) as f:
            f.write(html_content.encode('utf-8'))
            temp_path = f.name

        # Open the HTML file in the default browser
        webbrowser.open('file://' + os.path.abspath(temp_path))

        return temp_path

class CompilationManager:
    def __init__(self, fragment: Fragment):
        self.fragment = fragment

    def python_source(self):
        """Extract the source code that created this fragment from the caller's file."""
        try:
            # Start with the current frame and work backwards
            frame = inspect.currentframe()

            # Skip the first frame (this method)
            if frame:
                frame = frame.f_back

            relevant_frames = []
            max_frames_to_check = 10  # Limit how far back we go

            # Collect frames that might be relevant
            for i in range(max_frames_to_check):
                if frame is None:
                    break

                filename = frame.f_code.co_filename

                # Skip frames from internal Python modules
                if (not filename.startswith('<') and
                    'site-packages' not in filename and
                    '/lib/python' not in filename):

                    # Check whether this file imports from relationalai.early_access.builder
                    try:
                        with open(filename, 'r') as f:
                            content = f.read()
                            if ('from relationalai.early_access.builder' in content or
                                'import relationalai.early_access.builder' in content):
                                relevant_frames.append(frame)
                    except FileNotFoundError:
                        pass  # If we can't read the file, just continue

                frame = frame.f_back

            # If we found relevant frames, use the last one (furthest up the call stack)
            if relevant_frames:
                frame = relevant_frames[-1]
            else:
                # Fall back to the second frame if we didn't find any relevant ones
                frame = inspect.currentframe()
                for _ in range(2):
                    if frame is None:
                        return "Source code not available"
                    frame = frame.f_back

            if frame is None:
                return "Source code not available"

            # Get the filename and line number
            caller_info = inspect.getframeinfo(frame)
            filename = caller_info.filename
            lineno = caller_info.lineno

            # Read the source file
            if os.path.exists(filename):
                with open(filename, 'r') as f:
                    source_lines = f.readlines()

                # Extract a reasonable context around the line
                start_line = max(0, lineno - 20)
                end_line = min(len(source_lines), lineno + 20)

                # Try to find a complete test method or function
                # Look for the start of a function or method definition before the current line
                for i in range(lineno - 1, max(0, start_line - 10), -1):
                    if i >= len(source_lines):
                        continue
                    line = source_lines[i].strip()
                    if line.startswith('def ') or line.startswith('async def '):
                        start_line = i
                        break

                context_lines = source_lines[start_line:end_line]
                return ''.join(context_lines)

            return "Source file not found"
        except Exception as e:
            return f"Error extracting source code: {str(e)}"

    def ir(self):
        builder_model = self.fragment._model
        if not builder_model:
            raise ValueError("Fragment must be part of a model to compile to IR")
        ir_model = builder_model._to_ir()
        # TODO: do something with the fragment here
        # fragment_task = builder_model._compiler.compile_task(self.fragment)
        return ir_model

    def rel(self, ir_model=None):
        if ir_model is None:
            ir_model = self.ir()
        rel_compiler = RelCompiler()
        full_code = rel_compiler.compile(ir_model)
        return full_code

    def devmode(self, ir_model=None):
        if ir_model is None:
            ir_model = self.ir()
        devmode_compiler = DevModeCompiler()
        sql = devmode_compiler.compile(ir_model)
        return sql

    def compile(self):
        """
        Sends the various compiled forms of a fragment to the debugger.
        This function compiles the fragment to IR, Rel, and SQL, then logs
        the results to the debug system.

        Returns:
            Dictionary containing 'ir', 'rel', and 'sql' keys with the compiled outputs
        """
        try:
            ir_model = self.ir()
            ir_string = ir.node_to_string(ir_model)

            python_source = self.python_source()

            try:
                rel = self.rel(ir_model)
            except Exception:
                rel = None

            try:
                sql = self.devmode(ir_model)
            except Exception:
                sql = None

            return CompilationResult(python_source, ir_string, rel, sql)
        except Exception as e:
            raise e

    def display(self, compilation_result):
        """
        Displays the compilation results in a browser by creating a temporary HTML file.

        Args:
            compilation_result: Dictionary containing 'ir', 'rel', and 'sql' keys
                               as returned by the compile() function

        Returns:
            Path to the temporary HTML file
        """
        # Create HTML content with syntax highlighting using highlight.js
        html_content = """
        <!DOCTYPE html>
        <html>
        <head>
            <title>RAI Compilation Results</title>
            <link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.7.0/styles/default.min.css">
            <script src="https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.7.0/highlight.min.js"></script>
            <script src="https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.7.0/languages/sql.min.js"></script>
            <style>
                body {{ font-family: Arial, sans-serif; margin: 20px; }}
                .section {{ margin-bottom: 30px; }}
                h2 {{ color: #333; }}
                pre {{ background-color: #f5f5f5; padding: 10px; border-radius: 5px; overflow: auto; }}
                .error {{ color: red; }}
            </style>
        </head>
        <body>
            <h1>RAI Compilation Results</h1>

            <div class="section">
                <h2>IR (Intermediate Representation)</h2>
                <pre><code class="language-json">{}</code></pre>
            </div>

            <div class="section">
                <h2>Rel</h2>
                <pre><code class="language-sql">{}</code></pre>
            </div>

            <div class="section">
                <h2>SQL</h2>
                <pre><code class="language-sql">{}</code></pre>
            </div>

            <script>hljs.highlightAll();</script>
        </body>
        </html>
        """.format(
            html.escape(compilation_result.get('ir', 'Not available')) if compilation_result.get('ir') else '<span class="error">Error during IR compilation</span>',
            html.escape(compilation_result.get('rel', 'Not available')) if compilation_result.get('rel') else '<span class="error">Error during Rel compilation</span>',
            html.escape(compilation_result.get('sql', 'Not available')) if compilation_result.get('sql') else '<span class="error">Error during SQL compilation</span>'
        )

        # Create a temporary HTML file
        with tempfile.NamedTemporaryFile(suffix='.html', delete=False) as f:
            f.write(html_content.encode('utf-8'))
            temp_path = f.name

        # Open the HTML file in the default browser
        webbrowser.open('file://' + os.path.abspath(temp_path))

        return temp_path



from typing import Any

from relationalai.early_access.builder import Concept as QBConcept
from relationalai.early_access.builder.builder import python_types_to_concepts, Relationship as QBRelationship
from relationalai.early_access.dsl.orm.relationships import Relationship


class Concept(QBConcept):

    def __init__(self, model, name: str, extends: list[Any] = [], identify_by:dict[str, Any]={}):
        self._dsl_model = model
        # create an orm Relationship for each Concept to be able to refer to them in DSL model
        identify_args = {}
        if identify_by:
            for k, v in identify_by.items():
                if python_types_to_concepts.get(v):
                    v = python_types_to_concepts[v]
                if isinstance(v, Concept):
                    identify_args[k] = Relationship(self._dsl_model, f"{{{name}}} has {{{k}:{v._name}}}", short_name=k)
                elif isinstance(v, type) and issubclass(v, self._dsl_model.Enum): #type: ignore
                    identify_args[k] = Relationship(self._dsl_model, f"{{{name}}} has {{{k}:{v._concept._name}}}", short_name=k)
                elif isinstance(v, QBRelationship):
                    identify_args[k] = v
                else:
                    raise ValueError(f"identify_by must be either a Concept or Relationship: {k}={v}")
            # add constraints required for reference schema
            self._dsl_model._ref_scheme_constraints(*identify_args.values())
        super().__init__(name, extends, model.qb_model(), identify_args)
        self._dsl_model._add_concept(self)

    def identify_by(self, *relations:QBRelationship):
        super().identify_by(*relations)
        self._dsl_model._ref_scheme_constraints(*relations)

    def __eq__(self, other):
        return isinstance(other, Concept) and self._name == other._name

    def __hash__(self):
        return hash(self._name)

    def __repr__(self):
        return f"Concept({self._name})"
